# Use this node in the ast when you want to add some custom C++
#
# This is one mechanism to allow for a leaky abstraction.
import ast
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, cast

import func_adl_xAOD.common.cpp_types as ctyp
import func_adl_xAOD.common.statement as statements
from func_adl_xAOD.common.cpp_representation import (cpp_collection, cpp_value,
                                                     cpp_variable)
from func_adl_xAOD.common.cpp_vars import unique_name
from func_adl_xAOD.common.util_scope import gc_scope

# The list of methods and the re-write functions for them. Each rewrite function
# is called with the Call node, which includes arguments, names, etc. It should return
# None or a cpp_ast.
method_names = {}


class CPPCodeValue (ast.AST):
    r'''
    Represents a C++ bit of code that returns a value. Like a function call or a member call.
    Use the be-fore the wire visit phase of processing to look for a pattern that needs
    to generate AST code, like a method call. Then place this AST in place of the function.
    The back-end will then do the rendering useing the information included below.

    TODO: This should be a dataclass!
    '''

    def __init__(self):
        # Files that need to be included at the top of the generated C++ file
        self.include_files = []

        # List of link libraries
        self.link_libraries = []

        # Code that is run once at the start of each "event"
        self.initialization_code = []

        # Code that is run when the particular bit of code needs to be invoked (e.g. in the middle of a hot loop).
        # This is invoked in its own scope (between "{" and "}") so there are no variable collisions.
        self.running_code = []

        # The arguments to the function. These are "correctly" mapped into the argument values
        # that are passed to the function and then a text replacement is done in the code.
        self.args = []

        # Special replacement if this is a method call. A tuple. The first item is the string to be replaced in the
        # code. The second is the name against which we should be making the call (e.g. if j is the current jet variable,
        # the tuple might be ("obj", "j")).
        self.replacement_instance_obj = None

        # A string representing the result value. This must be a simple variable. It will get replaced
        # in all the code lines above.
        self.result: Optional[str] = None

        # Representation to use for the resulting variable. Includes C++ type information.
        # A lambda that takes teh scope as an argument and returns a cpp variable to hold things.
        self.result_rep: Optional[Callable[[gc_scope], cpp_variable]] = None

        # Instance declaration and initialization. The instance is initialized in the constructor.
        # The element is a tuple:(cpp_rep:instance_declaration, str: instance_initialization)
        self.fields = []


# Info used to build a code spec
@dataclass
class CPPCodeSpecification:
    # Name of the function or method
    name: str

    # List of include files that we should include accessing this guy
    include_files: List[str]

    # The names of the arguments (in python)
    arguments: List[str]

    # List of lines that will be used as the template for C++ code
    code: List[str]

    # The name of the result variable
    result: str

    # The type of the result variable, or if a collection, of the values
    cpp_return_type: str

    # True if this is a collection return type. In that case, the collection is expected
    # to obey vector like semantics
    cpp_return_is_collection: bool = False

    # The name of the object this method applies against
    method_object: Optional[str] = None

    # The name of the object if this is being used as a method (e.g. the `self` variable)
    instance_object: Optional[str] = None


def build_CPPCodeValue(spec: CPPCodeSpecification, call_node: ast.Call) -> ast.Call:
    '''
    Given the specification for a C++ code block, invoked as a function in our AST, replace
    the call node with a cpp spec callback AST so the C++ code is properly inserted into the
    call stream.


    Args:
        spec (CPPCodeSpecification): The specification, including the code, that we should insert at this call node
        call_node (ast.Call): The call node (with arguments) that we are going to replace with this
        C++ code.

    Raises:
        ValueError: Raised if something is wrong with the call site

    Returns:
        [type]: The C++ ast that can easily be emitted as code
    '''

    if len(call_node.args) != len(spec.arguments):
        raise ValueError(f"The call of {spec.name}({', '.join(spec.arguments)}) has insufficient arguments ({len(call_node.args)}).")

    if isinstance(call_node.func, ast.Attribute) and spec.method_object is None:
        raise ValueError(f"The {spec.name} is a function, but it is invoked like a method.")

    if isinstance(call_node.func, ast.Name) and spec.method_object is not None:
        raise ValueError(f"The {spec.name} is a method, but it is invoked like a function.")

    # Create an AST to hold onto all of this.
    r = CPPCodeValue()
    # We need TVector2 included here
    r.include_files += spec.include_files

    # We need all four arguments pushed through.
    r.args = spec.arguments

    # The code is three steps
    r.running_code += spec.code
    r.result = spec.result
    if spec.cpp_return_is_collection:
        r.result_rep = lambda scope: cpp_collection(unique_name(spec.name), scope=scope, collection_type=ctyp.collection(ctyp.terminal(spec.cpp_return_type)))  # type: ignore
    else:
        r.result_rep = lambda scope: cpp_variable(unique_name(spec.name), scope=scope, cpp_type=ctyp.terminal(spec.cpp_return_type))

    # If this is a mehtod, copy the info over to generate the obj reference.
    if spec.method_object is not None:
        r.replacement_instance_obj = (spec.method_object, call_node.func.value.id)  # type: ignore

    call_node.func = r  # type: ignore
    return call_node


class cpp_ast_finder(ast.NodeTransformer):
    r'''
    Look through the complete ast and replace method calls that are to a C++ plug in with a c++ ast
    node.
    '''
    def __init__(self, method_names: Dict[str, Callable[[ast.Call], ast.Call]]):
        self._method_names = method_names

    def try_call(self, name, node):
        'Try to use name to do the call. Returns (ok, result) monad'
        if name in self._method_names:
            cpp_call_ast = self._method_names[name](node)
            return (cpp_call_ast is not None, cpp_call_ast)
        return (False, None)

    def visit_Call(self, node):
        r'''
        Looking for a member call of a particular name. We rewrite that as
        another name.
        WARNING: currently the namespace is global, so the parent type doesn't matter!
        '''

        # Make sure all parts of this AST are visited properly before we attempt to
        # understand the call.
        self.generic_visit(node)

        # Examine the func to see if this is a member call.
        func = node.func
        if (type(func) is ast.Attribute) and (type(func.value) is ast.Name):
            ok, new_node = self.try_call(func.attr, node)
            if ok:
                return new_node
        elif type(func) is ast.Name:
            ok, new_node = self.try_call(func.id, node)
            if ok:
                return new_node

        return node


def process_ast_node(visitor, gc, call_node: ast.Call):
    r'''Inject the proper code into the output stream to deal with this C++ code.

    We expect this to be run on the back-end of the system.

    visitor - The node visitor that is converting the code into C++
    gc - the generated code object that we fill with actual code
    call_node - a Call ast node, with func being a CPPCodeValue.

    Result:
    representation - A value that represents the output
    '''

    # We write everything into a new scope to prevent conflicts. So we have to declare the result ahead of time.
    cpp_ast_node = cast(CPPCodeValue, call_node.func)
    result_rep = cpp_ast_node.result_rep(gc.current_scope())  # type: ignore

    gc.declare_variable(result_rep)

    # Include files and link libraries
    for i in cpp_ast_node.include_files:
        gc.add_include(i)
    for i in cpp_ast_node.link_libraries:
        gc.add_link_library(i)

    # Build the dictionary for replacement for the object we are calling
    # against, if any.
    repl_list = []
    if cpp_ast_node.replacement_instance_obj is not None:
        repl_list += [(cpp_ast_node.replacement_instance_obj[0], visitor.resolve_id(cpp_ast_node.replacement_instance_obj[1]).rep.as_cpp())]

    # Process the arguments that are getting passed to the function
    for arg, dest in zip(cpp_ast_node.args, call_node.args):
        rep = visitor.get_rep(dest)
        repl_list += [(arg, rep.as_cpp())]

    # Emit the statements.
    blk = statements.block()
    visitor._gc.add_statement(blk)

    for s in cpp_ast_node.running_code:
        l_s = s
        for src, dest in repl_list:
            l_s = l_s.replace(src, str(dest))
        blk.add_statement(statements.arbitrary_statement(l_s))

    # Emit the instance declaration and intialization code.
    for i in cpp_ast_node.fields:
        l_s = i[1]
        gc.declare_class_variable(i[0])
        for src, dest in repl_list:
            l_s = l_s.replace(src, str(dest))
        token_set_var = statements.set_var(i[0], cpp_value(l_s, None, None))
        gc.add_book_statement(token_set_var)

    # Set the result and close the scope
    assert cpp_ast_node.result is not None
    blk.add_statement(statements.set_var(result_rep, cpp_value(cpp_ast_node.result, gc.current_scope(), result_rep.cpp_type())))
    gc.pop_scope()

    return result_rep



import ast
from typing import Callable, Dict

import func_adl_xAOD.common.cpp_ast as cpp_ast
import func_adl_xAOD.common.cpp_types as ctyp
from func_adl_xAOD.common.cpp_representation import cpp_variable
from func_adl_xAOD.common.cpp_vars import unique_name


def isNonnullAst(call_node):
    r'''
    User is trying to test, on certian objects, if the object, when dereferenced, will be a null
    pointer. This is tricky for our data model because it treats an object as both a pointer and
    a object. Most of `func_adl_xAOD` is setup to deal with it as a pointer. This allows us to
    get around ths.
    '''

    if len(call_node.args) != 1:
        raise ValueError("Calling isNonnull(object) has incorrect number of arguments")

    # Create an AST to hold onto all of this.
    r = cpp_ast.CPPCodeValue()

    # We need all four arguments pushed through.
    r.args = ['cms_object']

    # The code is three steps
    r.running_code += ['auto result = (cms_object).isNonnull();']
    r.result = 'result'
    r.result_rep = lambda scope: cpp_variable(unique_name('is_non_null'), scope=scope, cpp_type=ctyp.terminal('bool'))

    call_node.func = r
    return call_node


def isNonnull(cms_object) -> bool:
    'See if dereferencing the cms object will return a null pointer or not'
    raise NotImplementedError('IsNonnull should never be called in python!')


def get_cms_functions() -> Dict[str, Callable[[ast.Call], ast.Call]]:
    return {
        'isNonnull': isNonnullAst
    }



# Drive the translate of the AST from start into a set of files, which one can then do whatever
# is needed to.
import ast
from func_adl_xAOD.common.event_collections import EventCollectionSpecification
from typing import Any, Callable, Dict, List
from func_adl_xAOD.common.meta_data import (
    InjectCodeBlock,
    JobScriptSpecification,
    process_metadata,
)
import os
import sys
from abc import ABC, abstractmethod
from collections import defaultdict, namedtuple
from pathlib import Path
import itertools

import func_adl_xAOD.common.cpp_ast as cpp_ast
import func_adl_xAOD.common.cpp_representation as crep
import jinja2
from func_adl.ast.aggregate_shortcuts import aggregate_node_transformer
from func_adl.ast.func_adl_ast_utils import change_extension_functions_to_calls
from func_adl.ast.function_simplifier import simplify_chained_calls
from func_adl.ast import extract_metadata
from func_adl_xAOD.common.ast_to_cpp_translator import query_ast_visitor
from func_adl_xAOD.common.cpp_functions import find_known_functions
from func_adl_xAOD.common.util_scope import top_level_scope

ExecutionInfo = namedtuple(
    "ExecutionInfo", "result_rep output_path main_script all_filenames"
)


class _cpp_source_emitter:
    r"""
    Helper class to emit C++ code as we go
    """

    def __init__(self):
        self._lines_of_query_code = []
        self._indent_level = 0

    def add_line(self, ll):
        "Add a line of code, automatically deal with the indent"
        if ll == "}":
            self._indent_level -= 1

        self._lines_of_query_code += [f"{'  ' * self._indent_level}{ll}"]

        if ll == "{":
            self._indent_level += 1

    def lines_of_query_code(self):
        return self._lines_of_query_code


# The following was copied from: https://www.oreilly.com/library/view/python-cookbook/0596001673/ch04s22.html
def _find(pathname: str, matchFunc=os.path.isfile):
    assert len(pathname) > 0
    all_paths = sys.path + ["/usr/local", os.path.dirname(__file__) + "/../../"]
    for dirname in all_paths:
        candidate = os.path.join(dirname, pathname)
        if matchFunc(candidate):
            return candidate
    all_dirs = ",".join(all_paths)
    raise RuntimeError(f"Can't find file '{pathname}'. Looked in {all_dirs}")


def _find_dir(path):
    return _find(path, matchFunc=os.path.isdir)  # type: ignore


def _is_format_request(a: ast.AST) -> bool:
    """Return true if the top level ast is a call to generate a ROOT file output.

    Args:
        ast (ast.AST): AST to check at the top level

    Returns:
        bool: True if the ast is not format agnostic.
    """
    if not isinstance(a, ast.Call):
        raise ValueError(
            f"A func_adl ast must start with a function call. This does not: {ast.dump(a)}"
        )
    if not isinstance(a.func, ast.Name):
        raise ValueError(
            f"A func_adl ast must start with a function call to something like Select or AsROOTTTree. This does not: {ast.dump(a)}"
        )
    return a.func.id == "ResultTTree"


class executor(ABC):
    def __init__(
        self,
        file_names: list,
        runner_name: str,
        template_dir_name: str,
        method_names: Dict[str, Callable[[ast.Call], ast.Call]],
        extended_md: Dict[str, Any] = {},
    ):
        self._file_names = file_names
        self._runner_name = runner_name
        self._template_dir_name = template_dir_name
        self._method_names = method_names
        self._job_option_blocks = []
        self._inject_blocks: List[InjectCodeBlock] = []
        self._extended_md = extended_md
        self._found_extended_md: Dict[str, List[Any]] = defaultdict(list)

    def extended_md(self, name: str) -> List[Any]:
        "Return all the extended for a given key"
        if name not in self._found_extended_md:
            return []
        return self._found_extended_md[name]

    def add_extended_md(self, extended_md: Dict[str, Any]):
        "Add extended metadata to the executor"
        self._extended_md.update(extended_md)

    def _copy_template_file(self, j2_env, info, template_file, final_dir: Path):
        "Copy a file to a final directory"
        j2_env.get_template(template_file).stream(info).dump(
            str(final_dir / template_file)
        )

    def reset(self):
        """Called before any work is done on a new ast. Resets object to ground zero.

        If anyone who inherits has global state, it should be reset here too.

        TODO: The fact this is here is because we are dealing with global variables, either
        the instance variables here or actual global state. We really should not have this.
        Some subtle bugs have already occurred because of this. Especially in running locally
        vs running up in ServiceX.
        """
        # Reset out object
        self._job_option_blocks = []
        self._inject_blocks = []
        self._extended_md = {}

        # Reset the type system
        import func_adl_xAOD.common.cpp_types as ctyp

        ctyp.g_method_type_dict = {}

    def apply_ast_transformations(self, a: ast.AST):
        r"""
        Run through all the transformations that we have on tap to be run on the client side.
        Return a (possibly) modified ast.
        """
        # Do tuple resolutions. This might eliminate a whole bunch fo code!
        a, meta_data = extract_metadata(a)
        cpp_functions = process_metadata(meta_data, self._extended_md)
        a = change_extension_functions_to_calls(a)
        a = aggregate_node_transformer().visit(a)
        a = simplify_chained_calls().visit(a)
        a = find_known_functions().visit(a)

        # Pull out any extended metadata
        extended_md_types = {type(x): k for k, x in self._extended_md.items()}
        for item in cpp_functions:
            if type(item) in extended_md_types.keys():
                self._found_extended_md[extended_md_types[type(item)]].append(item)

        # Any C++ custom code needs to be threaded into the ast
        method_names = dict(self._method_names)
        method_names.update(
            {
                md.name: (
                    (lambda call_node, md=md: cpp_ast.build_CPPCodeValue(md, call_node))
                    if isinstance(md, cpp_ast.CPPCodeSpecification)  # type: ignore
                    else self.build_collection_callback(md)
                )
                for md in cpp_functions
                if isinstance(
                    md, (cpp_ast.CPPCodeSpecification, EventCollectionSpecification)
                )
            }
        )
        a = cpp_ast.cpp_ast_finder(method_names).visit(a)

        # Save the injection blocks
        self._inject_blocks = [
            md for md in cpp_functions if isinstance(md, InjectCodeBlock)
        ]

        # Pull off any joboption blocks
        for m in cpp_functions:
            if isinstance(m, JobScriptSpecification):
                self._job_option_blocks.append(m)

        # And return the modified ast
        return a

    def _ib_fetch(self, name: str) -> List[str]:
        "Return items from inject code blocks"
        return list(itertools.chain(*[getattr(md, name) for md in self._inject_blocks]))

    @property
    def body_include_files(self) -> List[str]:
        "Return the list of include files for the query.cpp file"
        return self._ib_fetch("body_includes")

    @property
    def header_include_files(self) -> List[str]:
        "Return the list of include files for the header file"
        return self._ib_fetch("header_includes")

    @property
    def private_members(self) -> List[str]:
        "Return the list of private members for the query.cpp file"
        return self._ib_fetch("private_members")

    @property
    def instance_initialization(self) -> List[str]:
        "Return the list of initialziation of instances for the query.cpp file"
        return self._ib_fetch("instance_initialization")

    @property
    def ctor_lines(self) -> List[str]:
        "Return the Lines of C++ to add to the body of the constructor for the query.cpp file"
        return self._ib_fetch("ctor_lines")

    @property
    def link_libraries(self) -> List[str]:
        "Return the items to add to the `CMake LINK_LIBRARIES` list"
        return self._ib_fetch("link_libraries")

    @property
    def initialize_lines(self) -> List[str]:
        "Return the list of include files for the query.cpp file"
        return self._ib_fetch("initialize_lines")

    @abstractmethod
    def build_collection_callback(
        self, metadata: EventCollectionSpecification
    ) -> Callable[[ast.Call], ast.Call]:
        """Given the specification for a collection, build the callback that will replace the AST properly
        when it comes time. These collections are things like Jets, etc., and all off the top level event.
        """

    @abstractmethod
    def get_visitor_obj(self) -> query_ast_visitor:
        """Subclass should return a query ast visitor for the flavor of C++ backend
        implemented.

        Returns:
            query_ast_visitor: The ast visitor that can be used to convert the ast into
            code.
        """

    def add_to_replacement_dict(self) -> Dict[str, Any]:
        """Subclasses can over ride this to add new items to the template
        replacement dict

        Returns:
            Dict[str, Any]: New items to add to the replacement dict
        """
        return {}

    def write_cpp_files(self, ast: ast.AST, output_path: Path) -> ExecutionInfo:
        r"""
        Given the AST generate the C++ files that need to run. Return them along with
        the input files.
        """

        # Find the base file dataset and mark it.
        from func_adl import find_EventDataset

        file = find_EventDataset(ast)
        iterator = crep.cpp_variable(
            "bogus-do-not-use", top_level_scope(), cpp_type=None
        )
        crep.set_rep(file, crep.cpp_sequence(iterator, iterator, top_level_scope()))

        # Visit the AST to generate the code structure and find out what the
        # result is going to be.
        qv = self.get_visitor_obj()
        result_rep = qv.get_rep(ast) if _is_format_request(ast) else qv.get_as_ROOT(ast)

        # Emit the C++ code into our dictionaries to be used in template generation below.
        query_code = _cpp_source_emitter()
        qv.emit_query(query_code)
        book_code = _cpp_source_emitter()
        qv.emit_book(book_code)
        class_decl_code = qv.class_declaration_code()
        includes = qv.include_files() + self.body_include_files
        link_libraries = qv.link_libraries() + self.link_libraries

        # The replacement dict to pass to the template generator can now be filled
        info = {}
        info["query_code"] = query_code.lines_of_query_code()
        info["class_decl"] = class_decl_code
        info["book_code"] = book_code.lines_of_query_code()
        info["body_include_files"] = includes
        info["header_include_files"] = self.header_include_files
        info["private_members"] = self.private_members
        info["instance_initialization"] = self.instance_initialization
        info["initialize_lines"] = self.initialize_lines
        info["ctor_lines"] = self.ctor_lines
        info["link_libraries"] = link_libraries
        info.update(self.add_to_replacement_dict())

        # We use jinja2 templates. Write out everything.
        template_dir = _find_dir(self._template_dir_name)
        j2_env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_dir))

        for file_name in self._file_names:
            self._copy_template_file(j2_env, info, file_name, output_path)

        (output_path / self._runner_name).chmod(0o755)

        # Reset our object for the next call (e.g. reset global state)
        self.reset()

        # Build the return object.
        return ExecutionInfo(
            result_rep, output_path, self._runner_name, self._file_names
        )



from typing import Union
import func_adl_xAOD.common.cpp_types as ctyp
from func_adl_xAOD.common.event_collections import (
    EventCollectionSpecification, event_collection_coder,
    event_collection_collection_container, event_collection_container)


class atlas_xaod_event_collection_container(event_collection_container):
    def __init__(self, type_name: Union[str, ctyp.CPPParsedTypeInfo], p_depth: int = 1):
        super().__init__(type_name, p_depth=p_depth)

    def __str__(self):
        return f"const {self.type} *"


class atlas_xaod_event_collection_collection(event_collection_collection_container):
    def __init__(self, type_name: Union[str, ctyp.CPPParsedTypeInfo],
                 element_name: Union[str, ctyp.CPPParsedTypeInfo],
                 p_depth_type: int = 1, p_depth_element: int = 1):
        super().__init__(type_name, element_name, p_depth_element=p_depth_element, p_depth_type=p_depth_type)

    def __str__(self):
        return f"const {self.type}*"


# all the collections types that are available. This is required because C++
# is strongly typed, and thus we have to transmit this information.
atlas_xaod_collections = [
    EventCollectionSpecification('atlas', 'Jets',
                                 ['xAODJet/JetContainer.h'],
                                 atlas_xaod_event_collection_collection('xAOD::JetContainer', 'xAOD::Jet'),
                                 ['xAODJet'],
                                 ),
    EventCollectionSpecification('atlas', 'Tracks',
                                 ['xAODTracking/TrackParticleContainer.h'],
                                 atlas_xaod_event_collection_collection('xAOD::TrackParticleContainer', 'xAOD::TrackParticle'),
                                 ['xAODTracking'],
                                 ),
    EventCollectionSpecification('atlas', 'EventInfo',
                                 ['xAODEventInfo/EventInfo.h'],
                                 atlas_xaod_event_collection_container('xAOD::EventInfo'),
                                 ['xAODEventInfo'],
                                 ),
    EventCollectionSpecification('atlas', 'TruthParticles',
                                 ['xAODTruth/TruthParticleContainer.h', 'xAODTruth/TruthParticle.h', 'xAODTruth/TruthVertex.h'],
                                 atlas_xaod_event_collection_collection('xAOD::TruthParticleContainer', 'xAOD::TruthParticle'),
                                 ['xAODTruth'],
                                 ),
    EventCollectionSpecification('atlas', 'Electrons',
                                 ['xAODEgamma/ElectronContainer.h', 'xAODEgamma/Electron.h'],
                                 atlas_xaod_event_collection_collection('xAOD::ElectronContainer', 'xAOD::Electron'),
                                 ['xAODEgamma'],
                                 ),
    EventCollectionSpecification('atlas', 'Muons',
                                 ['xAODMuon/MuonContainer.h', 'xAODMuon/Muon.h'],
                                 atlas_xaod_event_collection_collection('xAOD::MuonContainer', 'xAOD::Muon'),
                                 ['xAODMuon'],
                                 ),
    EventCollectionSpecification('atlas', 'MissingET',
                                 ['xAODMissingET/MissingETContainer.h', 'xAODMissingET/MissingET.h'],
                                 atlas_xaod_event_collection_collection('xAOD::MissingETContainer', 'xAOD::MissingET'),
                                 ['xAODMissingET'],
                                 ),
]


def define_default_atlas_types():
    'Define the default atlas types'
    ctyp.add_method_type_info("xAOD::TruthParticle", "prodVtx", ctyp.terminal('xAODTruth::TruthVertex', p_depth=1))
    ctyp.add_method_type_info("xAOD::TruthParticle", "decayVtx", ctyp.terminal('xAODTruth::TruthVertex', p_depth=1))
    ctyp.add_method_type_info("xAOD::TruthParticle", "parent", ctyp.terminal('xAOD::TruthParticle', p_depth=1))
    ctyp.add_method_type_info("xAOD::TruthParticle", "child", ctyp.terminal('xAOD::TruthParticle', p_depth=1))


class atlas_event_collection_coder(event_collection_coder):
    def get_running_code(self, container_type: event_collection_container) -> list:
        return [f'{container_type} result = 0;',
                'ANA_CHECK (evtStore()->retrieve(result, collection_name));']



from typing import Dict, List

import func_adl_xAOD.common.cpp_types as ctyp

_type_priority: Dict[str, int] = {
    'int': 0,
    'float': 1,
    'double': 2,
}


def most_accurate_type(type_list: List[ctyp.terminal]) -> ctyp.terminal:
    '''
    Return the best type we can for most accurate sum. So if one is floating or double, return that.

    We only deal with terminals we know about.
    '''
    assert len(type_list) > 0, 'Internal error - looking for best types in null list'
    assert all(t.type in _type_priority for t in type_list), \
        f'Not all types ({", ".join(t.type for t in type_list)}) are known (known: {", ".join(_type_priority.keys())})'

    ordered = sorted(type_list, key=lambda t: _type_priority[t.type], reverse=True)
    return ordered[0]



# Scope related utilities
# see https://stackoverflow.com/questions/33533148/how-do-i-specify-that-the-return-type-of-a-method-is-the-same-as-the-class-itsel
# for info on this next line. Already looking forward to python 4...
import copy


def top_level_scope():
    '''
    Returns a top level scope. Basically, the class level.
    '''
    return gc_scope_top_level()


class gc_scope:
    'Internal class to track the scope of a statement or variable.'
    def __init__(self, scope_stack):
        self._scope_stack = copy.copy(scope_stack)

    def __getitem__(self, key: int):
        '''
        Return a new scope, some number "up" from where we are now. This uses standard
        array slicing in python. If you do 0 you'll get back the top level. If you do -1
        you will get back everything but the last thing. -2 last two things, etc.
        '''
        if len(self._scope_stack[:key]) == 0:
            raise RuntimeError("Winding up at the top level scope is not yet supported")

        return gc_scope(self._scope_stack[:key])

    def frame_statements(self, key):
        'Return the nth frame block. -1 means the last one, 0 means the deepest (top) one.'
        return self._scope_stack[key]

    def declare_variable(self, var) -> None:
        'Declare a class at the scope level'
        self._scope_stack[-1].declare_variable(var)

    def starts_with(self, c):
        '''
        Return true if the scope c matches the first part of our scope. False otherwise.
        '''
        if c.is_top_level() and self.is_top_level():
            return True
        if c.is_top_level():
            return True
        if self.is_top_level():
            return False

        if len(c._scope_stack) > len(self._scope_stack):
            return False

        return all([a is b for a, b in zip(self._scope_stack, c._scope_stack[:len(self._scope_stack)])])

    def is_top_level(self):
        return False


class gc_scope_top_level:
    def is_top_level(self):
        return True

    def __getitem__(self, key: int) -> gc_scope:
        'Calling this would be an internal error'
        raise NotImplementedError()

    def starts_with(self, c):
        'Starts with can only be true for top level if the other guy is top level'
        return type(c) is gc_scope_top_level


def deepest_scope(v1, v2):
    '''
    Returns the variable that is at the deepest scope. If they are equal in scope, then return v1.
    '''
    s1 = v1.scope()
    s2 = v2.scope()
    if not s2.starts_with(s1):
        return v1
    if s1.starts_with(s2):
        return v1
    return v2



# Some math utilities
import ast
from typing import Callable, Dict
import func_adl_xAOD.common.cpp_ast as cpp_ast


DeltaRSpec = cpp_ast.CPPCodeSpecification(
    'DeltaR',
    ['TVector2.h', 'math.h'],
    ['eta1', 'phi1', 'eta2', 'phi2'],
    [
        'auto d_eta = eta1 - eta2;',
        'auto d_phi = TVector2::Phi_mpi_pi(phi1-phi2);',
        'auto result = sqrt(d_eta*d_eta + d_phi*d_phi);'
    ],
    'result',
    'double'
)


def DeltaR(eta1, phi1, eta2, phi2) -> float:
    'Calculate the DeltaR between two eta,phi specified vectors'
    raise NotImplementedError('DeltaR should never be called in python!')


def get_math_methods() -> Dict[str, Callable[[ast.Call], ast.Call]]:
    return {
        'DeltaR': lambda call_node: cpp_ast.build_CPPCodeValue(DeltaRSpec, call_node)
    }



#!/usr/bin/env python

import FWCore.ParameterSet.Config as cms  # type: ignore
import os

process = cms.Process("Demo")

process.load("FWCore.MessageService.MessageLogger_cfi")

process.maxEvents = cms.untracked.PSet(input=cms.untracked.int32(-1))

filelistPath = "filelist.txt"
fileNames = tuple([f"file:{line}" for line in open(filelistPath, "r").readlines()])

process.source = cms.Source(
    "PoolSource",
    # replace 'myfile.root' with the source file you want to use
    fileNames=cms.untracked.vstring(*fileNames),
)

process.demo = cms.EDAnalyzer("Analyzer")

output_file = os.environ["CMS_OUTPUT_FILE"]

process.TFileService = cms.Service("TFileService", fileName=cms.string(output_file))

process.p = cms.Path(process.demo)



from __future__ import annotations

import copy
from dataclasses import dataclass
import logging
from typing import Dict, List, Optional, Union


@dataclass
class CPPParsedTypeInfo:
    """
    A parsed type, with the type and whether it's a pointer.
    """

    # The type name (`int`, `vector<float`, etc.)
    name: str

    # Pointer, and how many (2 for `int**`, 0 for `int`, etc.)
    pointer_depth: int

    # True if this is a const variable
    is_const: bool = False

    def __str__(self):
        return self.name + "*" * self.pointer_depth


def parse_type(t_name: str) -> CPPParsedTypeInfo:
    """Convert a type name string into info for a type

    Args:
        t_name (str): The type name (`float`, `float*`)

    Returns:
        CPPParsedTypeInfo: Parsed info from the type
    """
    ptr_depth = 0
    while True:
        t_name = t_name.strip()
        if t_name.endswith("*"):
            ptr_depth += 1
            t_name = t_name[:-1]
        else:
            break

    if t_name.startswith("const "):
        is_const = True
        t_name = t_name[6:]
    else:
        is_const = False

    return CPPParsedTypeInfo(t_name, ptr_depth, is_const)


class terminal:
    "Represents something we cannot see inside, like float, or int, or bool"

    def __init__(
        self,
        t: Union[str, CPPParsedTypeInfo],
        p_depth: int = 0,
        is_const: bool = False,
        tree_type: Optional[str] = None,
    ):
        """Create a terminal type - a type that we do not need to see inside

        * int
        * float
        * MyOwnStruct

        Use `p_depth` to represent how indirect to get down to the type (int* is 1, for
        example).)

        Args:
            t (str|CPPParsedTypeInfo): The type to represent
            p_depth (int): How many levels of indirection to get to the type
            is_const (bool): Whether this is a const type
            tree_type (str): Type to send back as a leaf.
        """
        if isinstance(t, CPPParsedTypeInfo):
            self._type = t.name
            self._p_depth = t.pointer_depth
            self._is_const = t.is_const
        else:
            self._type = t
            self._p_depth = p_depth
            self._is_const = is_const
        self._tree_type = tree_type

    def __str__(self):
        c_str = "const " if self.is_const else ""
        return f"{c_str}{self.type}{'*' * self._p_depth}"

    @property
    def is_a_pointer(self) -> bool:
        "Returns true if this terminal is a pointer"
        return self._p_depth > 0

    @property
    def p_depth(self) -> int:
        "Return how many levels of indirection to get to the type"
        return self._p_depth

    @property
    def is_const(self) -> bool:
        "Returns true if this terminal is a const type"
        return self._is_const

    def default_value(self):
        raise NotImplementedError()

    @property
    def type(self) -> str:
        return self._type

    @property
    def tree_type(self) -> terminal:
        return (
            self
            if self._tree_type is None
            else terminal(
                self._tree_type, p_depth=self._p_depth, is_const=self._is_const
            )
        )

    def get_dereferenced_type(self) -> terminal:
        "Type after dereferencing it once. Will throw if this type cannot be dereferenced"
        if self._p_depth == 0:
            raise RuntimeError(f"Cannot dereference type {self}")

        # Do deep copy because this needs to work in subclasses.
        new_t = copy.copy(self)
        new_t._p_depth -= 1
        return new_t


class terminal_enum_value(terminal):
    def __init__(self, en_info: ENumInfo):
        """Create a terminal that is an enum value.

        Args:
            t (str): The enum value as a type
        """
        super().__init__(str(en_info))


class collection(terminal):
    "Represents a collection/list/vector of the same type"

    def __init__(
        self,
        element_type: terminal,
        array_type: Optional[Union[str, CPPParsedTypeInfo]] = None,
        p_depth: int = 0,
    ):
        """Create a collection type, like `vector<float>`.

        Args:
            element_type (terminal): The element type, like a `terminal` of `float`.
            array_type (Optional[Union[str, CPPParsedTypeInfo]], optional): The type of the array. Defaults to None. Everything
                    is lifted from `array_type` if it is a `CPPParsedTypeInfo`.
            p_depth (int, optional): If the array type is a pointer or not. Defaults to 0. Ignored if `array_type` is `CPPParsedTypeInfo`.
        """
        if array_type is None:
            super().__init__(f"std::vector<{element_type}>", p_depth=p_depth)
        elif isinstance(array_type, CPPParsedTypeInfo):
            super().__init__(array_type)
        else:
            super().__init__(array_type, p_depth=p_depth)

        # And the element type we are representing
        self._element_type = element_type

    def token_type(self):
        # Return the type of the token. This token can be used to get data via
        # functions like getByToken()
        return None

    # TODO: Turn into a property
    @property
    def element_type(self) -> terminal:
        "The type of element that this collection holds"
        return self._element_type


###########################
# Manage types


@dataclass
class MethodInvokeInfo:
    "Method invocation info"

    # The return type
    r_type: terminal

    # Number of dereferences to apply to the calling object. Normally zero.
    # 0: obj.method()
    # 1: obj->method(), etc.
    deref_depth: int


g_method_type_dict: Dict[str, Dict[str, MethodInvokeInfo]] = {}


def add_method_type_info(
    type_string: str, method_name: str, t: terminal, deref_depth: int = 0
):
    """
    Define a return type for a method

    type_string         String of the object the method is calling against
    method_name         Name of the object
    t                   The type (terminal, collection, etc.) of return type
    """
    if type_string not in g_method_type_dict:
        g_method_type_dict[type_string] = {}
    g_method_type_dict[type_string][method_name] = MethodInvokeInfo(t, deref_depth)


def method_type_info(type_string: str, method_name: str) -> Optional[MethodInvokeInfo]:
    """
    Return the type of the method's return value
    """
    if type_string not in g_method_type_dict:
        return None
    if method_name not in g_method_type_dict[type_string]:
        return None
    return g_method_type_dict[type_string][method_name]


@dataclass
class ENumInfo:
    "Information about an enum"

    # The name of the enum
    name: str

    # List of the value names it can take on
    values: List[str]

    # The namespace we are hosted in
    ns: NameSpaceInfo

    # The full name
    @property
    def full_name(self) -> str:
        return f"{self.ns.full_name}.{self.name}"

    def __str__(self) -> str:
        return self.full_name

    def value_as_cpp(self, value: str) -> str:
        """Returns the C++ string for this value. It is assumed value
        is part of our value list.

        Args:
            value (str): The value we should use.

        Returns:
            str: The C++ specification for the type.
        """
        return f"{self.ns.full_name}::{value}".replace(".", "::")


class NameSpaceInfo:
    "Information about a namespace"

    def __init__(self, name, hosting_ns: Optional[NameSpaceInfo]):
        self.ns_name = name
        self.names_spaces = {}
        self.enums = {}
        self.parent_ns = hosting_ns

    ns_name: str

    names_spaces: Dict[str, NameSpaceInfo]

    enums: Dict[str, ENumInfo]

    parent_ns: Optional[NameSpaceInfo] = None

    @property
    def full_name(self) -> str:
        "The full name of the namespace"
        return (
            f"{self.parent_ns.full_name}.{self.ns_name}"
            if self.parent_ns is not None
            else self.ns_name
        )

    def get_ns(self, ns_name: str) -> Optional[NameSpaceInfo]:
        "Return the namespace with the given name, or None if it does not exist"
        return self.names_spaces.get(ns_name, None)

    def get_enum(self, enum_name: str) -> Optional[ENumInfo]:
        "Return the enum with the given name, or None if it does not exist"
        return self.enums.get(enum_name, None)

    def __str__(self) -> str:
        return self.full_name


# The top level namespaces we know about
g_toplevel_ns: Dict[str, NameSpaceInfo] = {}


def get_toplevel_ns(ns_name: str) -> Optional[NameSpaceInfo]:
    """
    Return the AST for a top level namespace. This is used to resolve
    the type of a method call.
    """
    return g_toplevel_ns.get(ns_name, None)


def define_ns(ns_name: str) -> NameSpaceInfo:
    """Define a namespace.

    `ns_name` can be in the form:
        - `bogus` - defines a single level namespace
        - 'bogus.sub' - creates we `NameSpaceInfo` objects, placing the one for
          `sub` inside the one for `bogus`, which is places in the global ns dict.

    If the namespace is already defined, this function will return the existing
    namespace.

    Args:
        ns_name (str): Name of the namespace to define

    Returns:
        NameSpaceInfo: The defined namespace.
    """
    parts = ns_name.split(".")
    v = g_toplevel_ns.get(parts[0], None)
    if v is None:
        v = NameSpaceInfo(parts[0], None)
        g_toplevel_ns[parts[0]] = v
    for p in parts[1:]:
        w = v.get_ns(p)
        if w is None:
            w = NameSpaceInfo(p, v)
            v.names_spaces[p] = w
        v = w
    return v


def define_enum(ns_name: str, enum_name: str, enum_values: List[str]) -> ENumInfo:
    """Define a new enum.

    The enum is defined in the `ns_name` namespace, with the name `enum_name` and
    `enum_values`. The newly created `ENumInfo` object is returned.

    It is an error to try to create the same `enum` twice.

    Args:
        ns_name (str): The name of the namespace (can contain `.`)
        enum_name (str): The name of the enum
        enum_values (List[str]): The list of names the enum is allowed.

    Returns:
        ENumInfo: _description_
    """
    ns = define_ns(ns_name)
    if enum_name in ns.enums:
        logging.getLogger(__name__).info(
            f"Enum {enum_name} already defined in namespace {ns_name}"
        )
        return ns.enums[enum_name]

    e = ENumInfo(enum_name, enum_values, ns)
    ns.enums[enum_name] = e

    return e



from func_adl_xAOD.common.ast_to_cpp_translator import query_ast_visitor
from func_adl_xAOD.common.statement import book_ttree, ttree_fill


class book_xaod_ttree(book_ttree):
    "Book an ATLAS TTree for writing out. Meant to be in the Book method"

    def __init__(self, tree_name, leaves):
        super().__init__(tree_name, leaves)

    def emit(self, e):
        "Emit the book statement for a tree"
        e.add_line(
            f'ANA_CHECK (book (TTree ("{self._tree_name}", "My analysis ntuple")));'
        )
        e.add_line(f'auto myTree = tree ("{self._tree_name}");')
        for var_pair in self._leaves:
            e.add_line(f'myTree->Branch("{var_pair[0]}", &{var_pair[1].as_cpp()});')


class xaod_ttree_fill(ttree_fill):
    "Fill a ATLAS TTree"

    def __init__(self, tree_name):
        super().__init__(tree_name)

    def emit(self, e):
        e.add_line(f'tree("{self._tree_name}")->Fill();')


class atlas_xaod_query_ast_visitor(query_ast_visitor):
    r"""
    Drive the conversion to C++ from the top level query
    for ATLAS xAOD
    """

    def __init__(self):
        prefix = "atlas_xaod"
        super().__init__(prefix)

    def create_book_ttree_obj(self, tree_name: str, leaves: list) -> book_ttree:
        return book_xaod_ttree(tree_name, leaves)

    def create_ttree_fill_obj(self, tree_name: str) -> ttree_fill:
        return xaod_ttree_fill(tree_name)



import ast
from func_adl_xAOD.common.meta_data import generate_script_block
from func_adl_xAOD.atlas.xaod.event_collections import (
    atlas_event_collection_coder,
    atlas_xaod_collections,
    define_default_atlas_types,
)
from typing import Any, Callable, Dict
from func_adl_xAOD.common.event_collections import EventCollectionSpecification
from func_adl_xAOD.common.math_utils import get_math_methods
from func_adl_xAOD.atlas.xaod.jets import get_jet_methods
from func_adl_xAOD.atlas.xaod.query_ast_visitor import atlas_xaod_query_ast_visitor
from func_adl_xAOD.common.executor import executor


class atlas_xaod_executor(executor):
    def __init__(self, template_dir_name="func_adl_xAOD/template/atlas/r21"):
        file_names = [
            "ATestRun_eljob.py",
            "package_CMakeLists.txt",
            "query.cxx",
            "query.h",
            "runner.sh",
        ]
        runner_name = "runner.sh"
        self._ecc = atlas_event_collection_coder()
        method_names: Dict[str, Callable[[ast.Call], ast.Call]] = {
            md.name: self.build_callback(self._ecc, md) for md in atlas_xaod_collections
        }
        method_names.update(get_jet_methods())
        method_names.update(get_math_methods())
        super().__init__(file_names, runner_name, template_dir_name, method_names)
        define_default_atlas_types()

    @staticmethod
    def build_callback(ecc, md):
        "Required due to by-reference lambda capture not working as expected in python"
        return lambda cd: ecc.get_collection(md, cd)

    def reset(self):
        """Reset our atlas default types"""
        super().reset()
        define_default_atlas_types()

    def get_visitor_obj(self):
        return atlas_xaod_query_ast_visitor()

    def add_to_replacement_dict(self) -> Dict[str, Any]:
        d1 = super().add_to_replacement_dict()

        # Combine metadata script blocks
        d = {"job_option_additions": generate_script_block(self._job_option_blocks)}
        d.update(d1)
        return d

    def build_collection_callback(
        self, metadata: EventCollectionSpecification
    ) -> Callable[[ast.Call], ast.Call]:
        """Build the AST analyzer callback for this collection.

        Args:
            metadata (EventCollectionSpecification): The metadata describing this collection

        Returns:
            Callable[[ast.Call], ast.Call]: Function that will implement what is needed to build the ast properly.
        """
        if metadata.backend_name != "atlas":
            raise ValueError(
                f'Attempt to create a collection from metadata for the {metadata.backend_name} backend; only "atlas" allowed.'
            )

        return lambda cd: self._ecc.get_collection(metadata, cd)



# Collected code to get collections from the event object
import ast
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Union

import func_adl_xAOD.common.cpp_ast as cpp_ast
import func_adl_xAOD.common.cpp_representation as crep
import func_adl_xAOD.common.cpp_types as ctyp
from func_adl_xAOD.common.cpp_vars import unique_name


class event_collection_container(ctyp.terminal, ABC):
    def __init__(self, type_name: Union[str, ctyp.CPPParsedTypeInfo], p_depth: int):
        super().__init__(type_name, p_depth=p_depth)

    @abstractmethod
    def __str__(self) -> str:
        '''Return the string representation of this event collection.
        Helpful for identifying it in ast dumps

        Returns:
            str: Description
        '''


class event_collection_collection_container(ctyp.collection, ABC):
    def __init__(self, type_name: Union[str, ctyp.CPPParsedTypeInfo],
                 element_name: Union[str, ctyp.CPPParsedTypeInfo],
                 p_depth_type: int, p_depth_element: int):
        super().__init__(
            ctyp.terminal(element_name, p_depth=p_depth_element),
            array_type=type_name, p_depth=p_depth_type
        )

    @abstractmethod
    def __str__(self) -> str:
        '''Return the string representation of this event collection.
        Helpful for identifying it in ast dumps

        Returns:
            str: Description
        '''


@dataclass
class EventCollectionSpecification:
    backend_name: str
    name: str

    # List of include files (e.g. ['xAODJet/Jet.h'])
    include_files: List[str]

    # The container information
    container_type: Union[event_collection_container, event_collection_collection_container]

    # List of libraries (e.g. ['xAODJet'])
    libraries: List[str]


class event_collection_coder(ABC):
    '''Contains code to generate collections accessing code in the backend
    '''
    def get_collection(self, md: EventCollectionSpecification, call_node: ast.Call):
        r'''
        Return a cpp ast for accessing the jet collection with the given arguments.
        '''
        # Get the name jet collection to look at.
        if len(call_node.args) != 1:
            raise ValueError(f"Calling {md.name} - only one argument is allowed")
        if not isinstance(call_node.args[0], ast.Str):
            raise ValueError(f"Calling {md.name} - only acceptable argument is a string")

        # Fill in the CPP block next.
        r = cpp_ast.CPPCodeValue()
        r.args = ['collection_name', ]
        r.include_files += md.include_files
        r.link_libraries += md.libraries

        self.get_running_code_CPPCodeValue(r, md)
        r.result = 'result'

        if issubclass(type(md.container_type), event_collection_collection_container):
            r.result_rep = lambda scope: crep.cpp_collection(unique_name(md.name.lower()), scope=scope, collection_type=md.container_type)  # type: ignore
        else:
            r.result_rep = lambda scope: crep.cpp_variable(unique_name(md.name.lower()), scope=scope, cpp_type=md.container_type)

        # Replace it as a function that is going to get called.
        call_node.func = r  # type: ignore
        return call_node

    def get_running_code_CPPCodeValue(self, cpv: cpp_ast.CPPCodeValue, md: EventCollectionSpecification):
        r'''
        Put the running code information stored in EventCollectionSpecification into CPPCodeValue. Can be
        overridden to store extra information(such as variable declarations).
        '''
        cpv.running_code = self.get_running_code(md.container_type)

    @abstractmethod
    def get_running_code(self, container_type: Union[event_collection_container, event_collection_collection_container]) -> List[str]:
        '''Return the code that will extract the collection from the event object

        Args:
            container_type (event_collection_container): The container to extract.

        Returns:
            List[str]: Lines of C++ code to execute to get this out.
        '''



import ast
from typing import Callable

from func_adl_xAOD.common.event_collections import EventCollectionSpecification
from func_adl_xAOD.common.executor import executor
from func_adl_xAOD.common.math_utils import get_math_methods

from .cms_functions import get_cms_functions
from .event_collections import (cms_aod_collections,
                                cms_event_collection_coder,
                                define_default_cms_types)
from .query_ast_visitor import cms_aod_query_ast_visitor


class cms_aod_executor(executor):
    def __init__(self):
        file_names = ['analyzer_cfg.py', 'Analyzer.cc', 'BuildFile.xml', "copy_root_tree.C", 'runner.sh']
        runner_name = 'runner.sh'
        template_dir_name = 'func_adl_xAOD/template/cms/r5'

        self._ecc = cms_event_collection_coder()
        method_names = {
            md.name: self.build_callback(self._ecc, md)
            for md in cms_aod_collections
        }
        method_names.update(get_math_methods())
        method_names.update(get_cms_functions())

        super().__init__(file_names, runner_name, template_dir_name, method_names)

        define_default_cms_types()

    def reset(self):
        '''Reset system to initial state
        '''
        super().reset()
        define_default_cms_types()

    @staticmethod
    def build_callback(ecc, md):
        'Required due to by-reference lambda capture not working as expected in python'
        return lambda cd: ecc.get_collection(md, cd)

    def get_visitor_obj(self):
        return cms_aod_query_ast_visitor()

    def build_collection_callback(self, metadata: EventCollectionSpecification) -> Callable[[ast.Call], ast.Call]:
        if metadata.backend_name != 'cms_aod':
            raise ValueError(f'Attempt to create a collection from metadata for the {metadata.backend_name} backend; only "atlas, cms_aod, or cms_miniaod" allowed.')

        return lambda cd: self._ecc.get_collection(metadata, cd)



# Code to aid with accessing jet collections
import ast

import func_adl_xAOD.common.cpp_ast as cpp_ast


def getAttribute(call_node: ast.Call):
    '''
    This is a common mistake as this is how it is written in C++. However, because we do
    not do templates, we need to do something more explicit. So, bomb out here.
    '''
    raise Exception("Do not call `getAttribute` - instead call `getAttributeFloat` or `getAttributeVectorFloat`.")


def get_jet_methods():
    get_attribute_float = cpp_ast.CPPCodeSpecification(
        name='getAttributeFloat',
        include_files=['vector'],
        arguments=['moment_name', ],
        code=['auto result = obj_j->getAttribute<float>(moment_name);'],
        result='result',
        cpp_return_type='float',
        method_object='obj_j',
        instance_object='xAOD::Jet_v1',
    )
    get_attribute_vector_float = cpp_ast.CPPCodeSpecification(
        name='getAttributeFloat',
        include_files=['vector'],
        arguments=['moment_name', ],
        code=['auto result = obj_j->getAttribute<std::vector<double>>(moment_name);'],
        result='result',
        cpp_return_type='double',
        cpp_return_is_collection=True,
        method_object='obj_j',
        instance_object='xAOD::Jet_v1',
    )
    return {
        'getAttribute': getAttribute,
        'getAttributeFloat': lambda call_node: cpp_ast.build_CPPCodeValue(get_attribute_float, call_node),
        'getAttributeVectorFloat': lambda call_node: cpp_ast.build_CPPCodeValue(get_attribute_vector_float, call_node),
    }



# Statements
from abc import ABC, abstractmethod  # For declaring abstract base class
from typing import Any

import func_adl_xAOD.common.cpp_representation as crep
import func_adl_xAOD.common.cpp_types as ctyp


class BlockException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class block:
    "This is a block of statements surrounded by a scoping (like open close bracket, for loop, etc.)"

    def __init__(self):
        self._statements = []
        self._variables = []
        self._rep_dict = {}

    def add_statement(self, s):
        "Add statement s to the list of statements"
        self._statements += [s]

    def declare_variable(self, n):
        "Declare a variable n, which is of type cpp_variable"
        self._variables += [n]

    def emit(self, e):
        "Render the block of code"
        e.add_line("{")
        for v in self._variables:
            init_value = (
                ""
                if not isinstance(v, crep.cpp_variable) or (v.initial_value() is None)
                else f" ({v.initial_value().as_cpp()})"  # type: ignore
            )
            e.add_line(f"{v.cpp_type()} {v.as_cpp()}{init_value};")
        for s in self._statements:
            s.emit(e)
        e.add_line("}")

    def get_rep(self, name: Any) -> Any:
        """Return the representation for some object. If we do not know its value
        then ask our parent for the value. Return None if we can't find it.

        Args:
            name:           Key for lookup

        Returns:
            None if there is nothing defined with that key in the hierarchy, or
            an actual value if there is.
        """
        if name in self._rep_dict:
            return self._rep_dict[name]
        return None

    def set_rep(self, name: Any, value: Any):
        """Defines the `value` for a lookup of `name` at any time in this block and
        below.

        Args:
            name:       The lookup key
            value       The value to be cached
        """
        if name in self._rep_dict:
            raise BlockException(
                f"Internal Error: Representation for {str(name)} already exists. Cannot set twice"
            )
        self._rep_dict[name] = value


class loop(block):
    "A for loop"

    def __init__(
        self, loop_var_rep: crep.cpp_value, collection_rep: crep.cpp_collection
    ):
        """
        Create a new implicit for loop statement. A new var is created, and the scope is set to
        be the one down from here.
        """
        block.__init__(self)
        self._collection = collection_rep
        self._loop_variable = loop_var_rep

    def emit(self, e):
        "Emit a for loop enclosed by a block of code"
        e.add_line(
            f"for (auto &&{self._loop_variable.as_cpp()} : {self._collection.as_cpp()})"
        )
        block.emit(self, e)


class iftest(block):
    "An if statement"

    def __init__(self, if_expr):
        block.__init__(self)
        self._expr = if_expr

    def emit(self, e):
        e.add_line(f"if ({self._expr.as_cpp()})")
        block.emit(self, e)


class elsephrase(block):
    "An else statement. Must come after you pop and if statement off"

    def __init__(self):
        block.__init__(self)

    def emit(self, e):
        "Emit an else statement"
        e.add_line("else")
        block.emit(self, e)


# By Inheriting from ABC we declare it as an Abstract Base Class
class book_ttree(ABC):
    "Book a TTree for writing out. Meant to be in the Book method"

    def __init__(self, tree_name, leaves):
        self._tree_name = tree_name
        self._leaves = leaves

    @abstractmethod
    def emit(self, e):
        # It is marked as an abstract method and will have to be
        # implemented before being used
        pass


class ttree_fill(ABC):
    "Fill a TTree"

    def __init__(self, tree_name):
        self._tree_name = tree_name

    @abstractmethod
    def emit(self, e):
        pass


class set_var:
    "Assigning a value to a variable"

    def __init__(self, target_var, value_var):
        r"""
        target_var, value_var: representations we will use
        """
        self._target = target_var
        self._value = value_var

    def emit(self, e):
        do_conversion = (
            self._target.has_cpp_type()
            and self._value.has_cpp_type()
            and self._target.cpp_type().type != self._value.cpp_type().type
        )
        if not do_conversion:
            e.add_line(f"{self._target.as_cpp()} = {self._value.as_cpp()};")
        else:
            e.add_line(
                f"{self._target.as_cpp()} = static_cast<{self._target.cpp_type().type}>({self._value.as_cpp()});"
            )


class push_back:
    "push a variable onto a vector"

    def __init__(self, target_collection: crep.cpp_value, value_var: crep.cpp_value):
        r"""
        target_col, value_var: representations we will use
        """
        self._target = target_collection
        self._value = value_var

    def emit(self, e):
        s_type = self._target.cpp_type()
        assert isinstance(s_type, ctyp.collection)
        do_conversion = (
            s_type.element_type is not None
            and self._value.has_cpp_type()
            and s_type.element_type.type != self._value.cpp_type().type
        )
        if not do_conversion:
            e.add_line(f"{self._target.as_cpp()}.push_back({self._value.as_cpp()});")
        else:
            e.add_line(
                f"{self._target.as_cpp()}.push_back(static_cast<{s_type.element_type.type}>({self._value.as_cpp()}));"
            )


class container_clear:
    "push a variable onto a vector"

    def __init__(self, collection):
        r"""
        target_col, value_var: representations we will use
        """
        self._collection = collection

    def emit(self, e):
        e.add_line(f"{self._collection.as_cpp()}.clear();")


class arbitrary_statement:
    "An arbitrary line of C++ code. Avoid if possible, as it makes analysis impossible"

    def __init__(self, line: str):
        self._line = line

    def emit(self, e):
        ll = self._line
        if not ll.endswith(";"):
            ll += ";"
        e.add_line(ll)



# Utility routines to help with variables

# A global counter to help with unique variable numbering.

unique_var_index = 0


def unique_name(name, is_class_var=False):
    r'''Will return a new C++ legal variable name that has been made unique with an index.

    name - Base name of the variable. For example, if it is "dude", then "dude7" might be the result.
    is_class_var - If true, this is intended to be defined at the class level. A "_" is added as prefix.

    returns:

    String of a new variable number.
    '''

    global unique_var_index
    v_name = ("_" if is_class_var else "") + name + str(unique_var_index)
    unique_var_index += 1
    return v_name



from copy import copy
from func_adl_xAOD.common.event_collections import EventCollectionSpecification
from func_adl_xAOD.common.cpp_ast import CPPCodeSpecification
from func_adl_xAOD.common.cpp_types import (
    add_method_type_info,
    collection,
    terminal,
    define_enum,
)
from func_adl_xAOD.common.cpp_types import CPPParsedTypeInfo, parse_type
from typing import Any, Dict, List, Union
from dataclasses import dataclass, field


@dataclass
class JobScriptSpecification:
    name: str
    script: List[str]
    depends_on: List[str]


@dataclass
class InjectCodeBlock:
    "Code to be directly injected into the hpp and cpp files"

    # Name of the code block
    name: str

    # Include files for the cpp code
    body_includes: List[str] = field(default_factory=list)

    # Include files for the hpp code
    header_includes: List[str] = field(default_factory=list)

    # Instance variable declarations
    private_members: List[str] = field(default_factory=list)

    # Instance variable ctor initializers
    instance_initialization: List[str] = field(default_factory=list)

    # Code lines to place in the constructor
    ctor_lines: List[str] = field(default_factory=list)

    # Lines to add to initialize statement
    initialize_lines: List[str] = field(default_factory=list)

    # Packages/Libraries to add to the CMake lib line
    link_libraries: List[str] = field(default_factory=list)


SpecificationTypes = Union[
    CPPCodeSpecification,
    EventCollectionSpecification,
    JobScriptSpecification,
    InjectCodeBlock,
]


def ok_to_add_code_block(spec, cpp_funcs: List[SpecificationTypes]) -> bool:
    """Check already added code blocks specs to see if there is one with the same name,
    and if so, if its contents are the same.

    Return ok if this is a unique code block
    Return false if this is a duplicate
    Throw an error if the name but not content matches.

    Args:
        spec (InjectCodeBlock): The block to be going after
        cpp_funcs ([type]): The list code blocks we should find a match in
    """
    for b in cpp_funcs:
        if isinstance(b, InjectCodeBlock) and b.name == spec.name:
            if b == spec:
                return False
            raise ValueError(
                f"Duplicate inject_code blocks with name {spec.name} that are not identical. Do not know which one to use! first: {b} second: {spec}"
            )
    return True


def process_metadata(
    md_list: List[Dict[str, Any]], extended_properties: Dict[str, Any] = {}
) -> List[SpecificationTypes]:
    """Process a list of metadata, in order.

    Args:
        md (List[Dict[str, str]]): The metadata to process

    Returns:
        List[X]: Metadata we've found
    """
    cpp_funcs: List[SpecificationTypes] = []
    for md in md_list:
        md_type = md.get("metadata_type")
        if md_type is None:
            raise ValueError(f"Metadata is missing `metadata_type` info ({md})")

        if md_type == "add_method_type_info":
            if "return_type" in md:
                # Single return type
                type_info = parse_type(md["return_type"])
                term = terminal(
                    type_info.name,
                    p_depth=type_info.pointer_depth,
                    tree_type=md.get("tree_type", None),
                )
            else:
                type_info_element = parse_type(md["return_type_element"])
                type_info_collection = (
                    parse_type(md["return_type_collection"])
                    if "return_type_collection" in md
                    else CPPParsedTypeInfo(f"std::vector<{type_info_element}>", 0)
                )
                term = collection(
                    terminal(type_info_element), array_type=type_info_collection
                )
            d_count = 0
            if "deref_count" in md:
                d_count = int(md["deref_count"])
            add_method_type_info(md["type_string"], md["method_name"], term, d_count)
        elif md_type == "inject_code":
            info = dict(md)
            del info["metadata_type"]
            if len(info) > 0:
                try:
                    spec = InjectCodeBlock(**info)
                except TypeError as e:
                    raise ValueError(f"Bad inject_code block item: {str(e)}")
                if ok_to_add_code_block(spec, cpp_funcs):
                    cpp_funcs.append(spec)
        elif md_type == "add_job_script":
            spec = JobScriptSpecification(
                name=md["name"],
                script=md["script"],
                depends_on=md.get("depends_on", []),
            )
            cpp_funcs.append(spec)
        elif md_type == "add_cpp_function":
            spec = CPPCodeSpecification(
                md["name"],
                md["include_files"],
                md["arguments"],
                md["code"],
                md["result_name"] if "result_name" in md else "result",
                md["return_type"],
                (
                    bool(md["return_is_collection"])
                    if "return_is_collection" in md
                    else False
                ),
                md["method_object"] if "method_object" in md else None,
                md["instance_object"] if "instance_object" in md else None,
            )
            cpp_funcs.append(spec)
        elif md_type == "add_atlas_event_collection_info":
            for k in md.keys():
                if k not in [
                    "metadata_type",
                    "name",
                    "include_files",
                    "container_type",
                    "element_type",
                    "contains_collection",
                    "link_libraries",
                ]:
                    raise ValueError(
                        f"Unexpected key {k} when declaring ATLAS collection metadata"
                    )
            if (md["contains_collection"] and "element_type" not in md) or (
                not md["contains_collection"] and "element_type" in md
            ):
                raise ValueError(
                    "In collection metadata, `element_type` must be specified if `contains_collection` is true and not if it is false"
                )

            from func_adl_xAOD.atlas.xaod.event_collections import (
                atlas_xaod_event_collection_collection,
                atlas_xaod_event_collection_container,
            )

            container_type = (
                atlas_xaod_event_collection_collection(
                    md["container_type"], md["element_type"]
                )
                if md["contains_collection"]
                else atlas_xaod_event_collection_container(md["container_type"])
            )
            link_libraries = [] if "link_libraries" not in md else md["link_libraries"]
            spec = EventCollectionSpecification(
                "atlas", md["name"], md["include_files"], container_type, link_libraries
            )
            cpp_funcs.append(spec)
        elif md_type == "add_cms_aod_event_collection_info":
            for k in md.keys():
                if k not in [
                    "metadata_type",
                    "name",
                    "include_files",
                    "container_type",
                    "element_type",
                    "contains_collection",
                    "element_pointer",
                ]:
                    raise ValueError(
                        f"Unexpected key {k} when declaring CMS AOD collection metadata"
                    )
            if (md["contains_collection"] and "element_type" not in md) or (
                not md["contains_collection"] and "element_type" in md
            ):
                raise ValueError(
                    "In collection metadata, `element_type` must be specified if `contains_collection` is true and not if it is false"
                )

            from func_adl_xAOD.cms.aod.event_collections import (
                cms_aod_event_collection_collection,
            )

            container_type = cms_aod_event_collection_collection(
                md["container_type"], md["element_type"]
            )

            spec = EventCollectionSpecification(
                "cms_aod", md["name"], md["include_files"], container_type, []
            )
            cpp_funcs.append(spec)
        elif md_type == "add_cms_miniaod_event_collection_info":
            for k in md.keys():
                if k not in [
                    "metadata_type",
                    "name",
                    "include_files",
                    "container_type",
                    "element_type",
                    "contains_collection",
                    "element_pointer",
                ]:
                    raise ValueError(
                        f"Unexpected key {k} when declaring CMS MiniAOD collection metadata"
                    )
            if (md["contains_collection"] and "element_type" not in md) or (
                not md["contains_collection"] and "element_type" in md
            ):
                raise ValueError(
                    "In collection metadata, `element_type` must be specified if `contains_collection` is true and not if it is false"
                )

            from func_adl_xAOD.cms.miniaod.event_collections import (
                cms_miniaod_event_collection_collection,
            )

            container_type = cms_miniaod_event_collection_collection(
                md["container_type"], md["element_type"]
            )

            spec = EventCollectionSpecification(
                "cms_miniaod", md["name"], md["include_files"], container_type, []
            )
            cpp_funcs.append(spec)
        elif md_type == "define_enum":
            define_enum(md["namespace"], md["name"], md["values"])
        elif md_type in extended_properties:
            r = copy(extended_properties[md_type])
            for k in (all_k for all_k in md.keys() if all_k != "metadata_type"):
                setattr(r, k, md[k])
            cpp_funcs.append(r)
        else:
            raise ValueError(f"Unknown metadata type ({md_type})")

    return cpp_funcs


def generate_script_block(blocks: List[JobScriptSpecification]) -> List[str]:
    """Returns the script block to insert into the job control.

    * Takes dependencies into account
    * Gets rid of any duplicates
    * Will combine any blocks with the same script text, but different dependencies.

    Args:
        blocks (List[JobScriptSpecification]): The list of, unordered, dependency blocks

    Returns:
        List[str]: The list of insertions to insert.
    """
    # Build the dependency graph and check for improper
    # duplications and dependencies.
    dependencies: Dict[str, List[str]] = {}
    block_lookup: Dict[str, JobScriptSpecification] = {}

    for b in blocks:
        if b.name not in dependencies:
            dependencies[b.name] = []
            block_lookup[b.name] = b
        else:
            if b.script != block_lookup[b.name].script:
                raise ValueError(
                    f'Duplicate metadata block "{b.name}", but blocks are not identical ({b.script} and {block_lookup[b.name].script} should be identical)!'
                )

        dependencies[b.name].extend(b.depends_on)

    for name, deps in dependencies.items():
        for d in deps:
            if d not in dependencies:
                raise ValueError(
                    f"Dependent metadata block {d} not found in sent metadata (from {name}!"
                )

    # Next, start from blocks that have no dependencies and work our way up.
    seen_blocks = set()
    script_text = []

    while len(seen_blocks) < len(dependencies):
        emitted = False
        for j in block_lookup.values():
            if j.name not in seen_blocks:
                if set(dependencies[j.name]) <= seen_blocks:
                    for ln in j.script:
                        script_text.append(ln)
                    seen_blocks.add(j.name)
                    emitted = True
        if not emitted:
            remaining_blocks = ", ".join((set(block_lookup.keys()) - seen_blocks))
            raise ValueError(
                f"There seems to be a metadata script block circular dependency ({remaining_blocks})"
            )

    return script_text



from typing import Union
import func_adl_xAOD.common.cpp_types as ctyp
from func_adl_xAOD.common.event_collections import (
    EventCollectionSpecification, event_collection_coder, event_collection_collection_container, event_collection_container)


# There is no use for a CMS single item collection - everything they have
# has multiple items in it. Copy from the ATLAS example to get this working correctly.


class cms_aod_event_collection_collection(event_collection_collection_container):
    def __init__(self, type_name: Union[str, ctyp.CPPParsedTypeInfo],
                 element_name: Union[str, ctyp.CPPParsedTypeInfo],
                 p_depth_type: int = 1, p_depth_element: int = 0):
        super().__init__(type_name, element_name, p_depth_element=p_depth_element, p_depth_type=p_depth_type)

    def __str__(self):
        return f"edm::Handle<{self.type}>"


# all the collections types that are available. This is required because C++
# is strongly typed, and thus we have to transmit this information.
cms_aod_collections = [
    EventCollectionSpecification('cms_aod', "Tracks",
                                 ['DataFormats/TrackReco/interface/Track.h',
                                  'DataFormats/TrackReco/interface/TrackFwd.h',
                                  'DataFormats/TrackReco/interface/HitPattern.h'
                                  ],
                                 cms_aod_event_collection_collection('reco::TrackCollection', 'reco::Track'),
                                 [],
                                 ),
    EventCollectionSpecification('cms_aod', "TrackMuons",
                                 ['DataFormats/MuonReco/interface/Muon.h',
                                  'DataFormats/MuonReco/interface/MuonFwd.h',
                                  'DataFormats/MuonReco/interface/MuonSelectors.h',
                                  'DataFormats/MuonReco/interface/MuonIsolation.h',
                                  'DataFormats/MuonReco/interface/MuonPFIsolation.h',
                                  'DataFormats/TrackReco/interface/Track.h',
                                  'DataFormats/TrackReco/interface/TrackFwd.h',
                                  'DataFormats/TrackReco/interface/HitPattern.h'
                                  ],
                                 cms_aod_event_collection_collection('reco::TrackCollection', 'reco::Track'),
                                 [],
                                 ),
    EventCollectionSpecification('cms_aod', "Muons",
                                 ['DataFormats/MuonReco/interface/Muon.h',
                                  'DataFormats/MuonReco/interface/MuonFwd.h',
                                  'DataFormats/MuonReco/interface/MuonSelectors.h',
                                  'DataFormats/MuonReco/interface/MuonIsolation.h',
                                  'DataFormats/MuonReco/interface/MuonPFIsolation.h'
                                  ],
                                 cms_aod_event_collection_collection('reco::MuonCollection', 'reco::Muon'),
                                 [],
                                 ),
    EventCollectionSpecification('cms_aod', "Vertex",
                                 ["DataFormats/VertexReco/interface/Vertex.h",
                                  "DataFormats/VertexReco/interface/VertexFwd.h"
                                  ],
                                 cms_aod_event_collection_collection('reco::VertexCollection', 'reco::Vertex', p_depth_element=0),
                                 [],
                                 ),
    EventCollectionSpecification('cms_aod', "GsfElectrons",
                                 ['DataFormats/EgammaCandidates/interface/GsfElectron.h',
                                  'DataFormats/GsfTrackReco/interface/GsfTrack.h',
                                  'DataFormats/GsfTrackReco/interface/GsfTrackFwd.h'
                                  ],
                                 cms_aod_event_collection_collection('reco::GsfElectronCollection', 'reco::GsfElectron'),
                                 [],
                                 ),
]


def define_default_cms_types():
    'Define the default cms types'
    ctyp.add_method_type_info("reco::Track", "hitPattern", ctyp.terminal('reco::HitPattern'))

    ctyp.add_method_type_info("reco::Muon", "globalTrack", ctyp.terminal('reco::Track', p_depth=1))
    ctyp.add_method_type_info("reco::Muon", "hitPattern", ctyp.terminal('reco::HitPattern'))
    ctyp.add_method_type_info("reco::Muon", "isPFIsolationValid", ctyp.terminal('bool'))
    ctyp.add_method_type_info("reco::Muon", "isPFMuon", ctyp.terminal('bool'))
    ctyp.add_method_type_info("reco::Muon", "pfIsolationR04", ctyp.terminal('reco::MuonPFIsolation'))

    ctyp.add_method_type_info("reco::GsfElectron", "gsfTrack", ctyp.terminal('reco::GsfTrack', p_depth=1))
    ctyp.add_method_type_info("reco::GsfElectron", "isEB", ctyp.terminal('bool'))
    ctyp.add_method_type_info("reco::GsfElectron", "isEE", ctyp.terminal('bool'))
    ctyp.add_method_type_info("reco::GsfElectron", "passingPflowPreselection", ctyp.terminal('bool'))
    ctyp.add_method_type_info("reco::GsfElectron", "superCluster", ctyp.terminal('reco::SuperClusterRef', p_depth=1))
    ctyp.add_method_type_info("reco::GsfElectron", "pfIsolationVariables", ctyp.terminal('reco::GsfElectron::PflowIsolationVariables'))

    ctyp.add_method_type_info("reco::GsfTrack", "trackerExpectedHitsInner", ctyp.terminal('reco::HitPattern'))  # reco::HitPattern is the expected return type


class cms_event_collection_coder(event_collection_coder):
    def get_running_code(self, container_type: event_collection_container) -> list:
        return [f'{container_type} result;',
                'iEvent.getByLabel(collection_name, result);']



#
# Read the submission directory as a command line argument. You can
# extend the list of arguments with your private ones later on.
# Set up (Py)ROOT.
import ROOT  # type: ignore
import optparse
from AnaAlgorithm.DualUseConfig import createAlgorithm  # type: ignore

parser = optparse.OptionParser()

ROOT.xAOD.Init().ignore()
parser.add_option('-s', '--submission-dir', dest='submission_dir',
                  action='store', type='string', default='submitDir',
                  help='Submission directory for EventLoop')
(options, args) = parser.parse_args()


# The sample handler is going to load the files form filelist.txt,
# in this context, it is an embarrassingly easy use of that object.
sh = ROOT.SH.SampleHandler()
sh.setMetaString('nc_tree', 'CollectionTree')
ROOT.SH.readFileList(sh, "ANALYSIS", "filelist.txt")
sh.printContent()

# Create an EventLoop job.
job = ROOT.EL.Job()
job.sampleHandler(sh)

{% for i in job_option_additions %}
{{i}}
{% endfor %}

# Create the algorithm's configuration.
alg = createAlgorithm('query', 'AnalysisAlg')
# later on we'll add some configuration options for our algorithm that go here

# Add our algorithm to the job
job.algsAdd(alg)
job.outputAdd(ROOT.EL.OutputStream('ANALYSIS'))

# Run the job using the direct driver.
driver = ROOT.EL.DirectDriver()
driver.submit(job, options.submission_dir)



from pathlib import Path
from typing import List, Optional, Union

from func_adl_xAOD.common.local_dataset import LocalDataset, docker_volume_info
from func_adl_xAOD.cms.aod.executor import cms_aod_executor
from func_adl_xAOD.common.executor import executor


class CMSRun1AODDataset(LocalDataset):
    '''A dataset running locally
    '''
    def __init__(self,
                 files: Union[Path, str, List[Path], List[str]],
                 docker_image: str = 'cmsopendata/cmssw_5_3_32',
                 docker_tag: str = 'conddb_20210705',
                 output_directory: Optional[Path] = None):
        '''Run on the given files

        Args:
            files (Path): Locally accessible files we are going to run on
            docker_image (str): The docker image name to run the executable
            docker_tag (str): The docker tag to use to run the executable
        '''
        super().__init__(files, docker_image, docker_tag, output_directory)

    def get_executor_obj(self) -> executor:
        '''Return the code that will actually generate the C++ we need to execute
        here.

        Returns:
            executor: Return the executor
        '''
        return cms_aod_executor()

    def docker_cache_volume(self) -> List[docker_volume_info]:
        return []



# Hold onto the generated code
from typing import List, Union

from func_adl_xAOD.common.statement import block
from func_adl_xAOD.common.util_scope import gc_scope, gc_scope_top_level


class generated_code:
    def __init__(self):
        self._block = block()
        self._book_block = block()
        self._class_vars = []
        self._scope_stack = (self._block,)
        self._include_files = []
        self._link_libraries = []

    def declare_class_variable(self, var):
        "Declare a variable as an instance of the query class. var must be a cpp_rep"
        self._class_vars += [var]

    def declare_variable(self, v):
        "Declare a variable at the current scope"
        self._scope_stack[-1].declare_variable(v)

    def get_rep(self, name):
        "Get a representation that has been defined on the stack"
        items = (
            rep
            for rep in (s.get_rep(name) for s in reversed(self._scope_stack))
            if rep is not None
        )
        return next(items, None)

    def set_rep(self, name, value):
        "Set a representation for later recall"
        self._scope_stack[-1].set_rep(name, value)

    def add_statement(self, st, below=None):
        """
        Add a statement. By default it is added to wherever the current
        cursor/stack is pointing (to the end of the blow). If we are adding a new
        block, then the cursor is update so future insertions to that block.

        st - The statement to add
        below - If not none, then the statement is added below the given statement,
                and everything that was below is put inside the statement here. The
                current point of insertion is not affected.
        """
        if below is None:
            self._scope_stack[-1].add_statement(st)
            if isinstance(st, block):
                self._scope_stack = self._scope_stack + (st,)
        else:
            if not isinstance(below, block):
                raise Exception(
                    "Internal Error: Can't a statement below a statement that isn't a scoping block."
                )
            if not isinstance(st, block):
                raise Exception(
                    "Internal Error: Can't a statement that isn't a scoping block."
                )
            for s in below._statements:
                st.add_statement(s)
            below._statements = []
            below.add_statement(st)

    def add_include(self, path):
        self._include_files += [path]

    def include_files(self):
        return self._include_files

    def add_link_library(self, library: str):
        self._link_libraries += [library]

    def link_libraries(self) -> List[str]:
        return self._link_libraries

    def pop_scope(self):
        self._scope_stack = self._scope_stack[:-1]

    def current_scope(self):
        "Return a token that can be later used to set the scoping"
        return gc_scope(self._scope_stack)

    def set_scope(self, scope_info: Union[gc_scope, gc_scope_top_level]):
        "Set the scope to a previously cached value"
        if scope_info is None:
            raise Exception("Scope can't be set to null")
        if scope_info.is_top_level():
            # Special case this guy as it is a unicorn.
            self._scope_stack = self._scope_stack[:1]
            return

        # Restore it to whatever it was.
        assert isinstance(scope_info, gc_scope)
        self._scope_stack = scope_info._scope_stack  # type: ignore

    def add_book_statement(self, st, below=None):
        self._book_block.add_statement(st)

    def emit_query_code(self, e):
        "Emit query code"
        self._block.emit(e)

    def emit_book_code(self, e):
        "Emit the book method code"
        self._book_block.emit(e)

    def class_declaration_code(self):
        "Return the class variable decls"
        s = []
        for v in self._class_vars:
            s += [f"{v.cpp_type()} {v.as_cpp()};\n"]

        return s



import ast
from dataclasses import dataclass
import logging
import shutil
import tempfile
from abc import ABC, abstractmethod
from collections.abc import Iterable
from pathlib import Path
from typing import Any, Callable, List, Optional, Union

import python_on_whales
from func_adl import EventDataset
from func_adl_xAOD.common.executor import executor
from func_adl_xAOD.common.result_ttree import cpp_ttree_rep
from python_on_whales import docker


@dataclass
class docker_volume_info:
    # Name of the volume in docker. Will be re-used.
    docker_name: str

    # The directory where the volume should be mounted
    mount_point: str


@dataclass
class DockerImageSpecification:
    "The image to run for docker"
    image: str


def _docker_volume_name(info: docker_volume_info) -> str:
    return "func_adl_" + info.docker_name


class LocalDataset(EventDataset, ABC):
    """A dataset running locally"""

    def __init__(
        self,
        files: Union[Path, str, List[Path], List[str]],
        docker_image: str,
        docker_tag: str,
        output_directory: Optional[Path] = None,
    ):
        """Run on the given files locally using a docker image to process them.

        When using `.value()` a list of `Path` objects is returned. All input files
        are combined into a single file. And the file is always called `ANALYSIS.root`

        NOTE:

        * The caller is responsible for the clean up of any output files that are
        generated. Use `output_directory` to control where they are written.
        * If you run multiple instances of these note that the output file is always
        called the same thing - so collisions are very likely!

        Args:
            files (Path): Locally accessible files we are going to run on
            docker_image (str): The docker image name to run the executable
            docker_tag (str): The docker tag to use to run the executable
            output_directory (Optional[Path], optional): The directory to write the output to.
                If `None` then will be written to the temp directory. Any directory passed in must
                already exist. Defaults to `None`.
        """
        super().__init__()

        if isinstance(files, str):
            f_list = [files]
        else:
            f_list = files if isinstance(files, Iterable) else [files]
        self.files = [Path(f) if not isinstance(f, Path) else f for f in f_list]

        if len(self.files) == 0:
            raise RuntimeError(
                "No files were given to the local dataset - need at least one good file"
            )

        self._docker_image = f"{docker_image}:{docker_tag}"

        for f in self.files:
            if not f.exists():
                raise FileNotFoundError(f"File {f} does not exist")

        assert tempfile.tempdir is not None
        self._output_directory = (
            output_directory if output_directory is not None else Path(tempfile.tempdir)
        )

        # Put everything into the ast so that we can safely be carried over qastle and used in
        # determining a hash key to see when things change.
        self.query_ast.args.append(ast.Str(s=f"{self._docker_image}"))  # type: ignore
        self.query_ast.args.append(ast.List(elts=[ast.Str(s=str(f)) for f in self.files]))  # type: ignore

    @abstractmethod
    def get_executor_obj(self) -> executor:
        """Return the code that will actually generate the C++ we need to execute
        here.

        Returns:
            executor: Return the executor
        """

    @abstractmethod
    def docker_cache_volume(self) -> List[docker_volume_info]:
        """Return info for a cache volume that can be mounted to the docker container.

        Used to cache downloaded files between runs (like calibration files).

        Returns:
            str: The volume string
        """

    async def execute_result_async(self, a: ast.AST, title: str) -> Any:
        """Take the `ast` and turn it into code and run it in docker, async.

        Args:
            a (ast.AST): The AST fo the query to run
            title (str): Title of the query

        Returns:
            Any: List of files
        """
        # Build everything in the local temp file directory.
        with tempfile.TemporaryDirectory() as local_run_dir_p:

            # Setup the local directory and make sure it is writeable
            local_run_dir = Path(local_run_dir_p)
            local_run_dir.chmod(0o777)

            # Get the files that we can run
            exe = self.get_executor_obj()
            exe.add_extended_md(
                {"docker": DockerImageSpecification(self._docker_image)}
            )
            f_spec = exe.write_cpp_files(
                exe.apply_ast_transformations(a), local_run_dir
            )

            # Get the image
            docker_image = self._docker_image
            md = exe.extended_md("docker")
            if len(md) > 0:
                docker_image = md[-1].image

            # Write out a file with the mapped in directories.
            # Until we better figure out how to deal with this, there are some restrictions
            # on file locations.
            datafile_dir: Optional[Path] = None
            with open(f"{local_run_dir}/filelist.txt", "w") as flist_out:
                for u in self.files:

                    ds_path = u.parent
                    datafile = u.name
                    flist_out.write(f"/data/{datafile}\n")
                    if datafile_dir is None:
                        datafile_dir = ds_path
                    else:
                        if ds_path != datafile_dir:
                            raise RuntimeError(
                                f"Data files must be from the same directory. Have seen {ds_path} and {datafile_dir} so far."
                            )

            # Build the docker command and run it.
            volumes_to_mount = [
                (f_spec.output_path, "/scripts", "ro"),
                (f_spec.output_path, "/results", "rw"),
                (datafile_dir, "/data/", "ro"),
            ]

            # Add any docker volumes in
            for v_info in self.docker_cache_volume():
                # Make sure the volume has been created
                v_name = _docker_volume_name(v_info)
                # if not docker.volume.exists(v_name):
                #     docker.volume.create(v_name)
                volumes_to_mount.append((v_name, v_info.mount_point))

            output: str = ""
            try:
                output_generator = docker.run(
                    docker_image,
                    [f"/scripts/{f_spec.main_script}"],
                    volumes=volumes_to_mount,
                    remove=True,
                    stream=True,
                )
                for stream_type, stream_content in output_generator:  # type: ignore
                    if stream_type == "stdout":
                        output += f"{stream_content.decode()}"
                    else:
                        output += f"(stderr) {stream_content.decode()}"
                self._dump_info(
                    logging.DEBUG,
                    output,
                    local_run_dir,
                    f_spec.main_script,
                    self._docker_image,
                )

            except python_on_whales.exceptions.DockerException as e:
                self._dump_info(
                    logging.ERROR,
                    output,
                    local_run_dir,
                    f_spec.main_script,
                    self._docker_image,
                )
                raise e

            # Now that we have run, we can pluck out the result.
            assert isinstance(f_spec.result_rep, cpp_ttree_rep), "Unknown return type"
            return [
                _extract_result_TTree(
                    f_spec.result_rep, local_run_dir, self._output_directory
                )
            ]

    def _dump_info(
        self,
        level,
        running_string: str,
        local_run_dir: Path,
        source_file_name: str,
        docker_image: str,
    ):
        """Dump the logging info from a docker run.

        Args:
            level ([type]): The logging level
            running_string (str): The string message from the run
        """
        lg = logging.getLogger(__name__)

        lg.log(level, f"Docker image and tag: {docker_image}")
        lg.log(level, "Docker Output: ")
        _dump_split_string(running_string, lambda line: lg.log(level, f"  {line}"))

        for file in local_run_dir.glob("*"):
            if file.is_file() and (file.suffix != ".root"):
                lg.log(level, f"{file.name}:")
                with file.open("r") as f:
                    _dump_split_string(
                        f.read(), lambda line: lg.log(level, f"  {line}")
                    )


def _dump_split_string(s: str, dump: Callable[[str], None]):
    for ll in s.split("\n"):
        dump(ll)


def _extract_result_TTree(rep: cpp_ttree_rep, run_dir, output_dir: Path):
    """Copy the final file into a place that is "safe", and return that as a path.

    The reason for this is that the temp directory we are using is about to be deleted!

    Args:
        rep (cpp_base_rep): The representation of the final result
        run_dir ([type]): Directory where it ran

    Raises:
        Exception: [description]
    """
    current_path = run_dir / rep.filename
    new_path = output_dir / rep.filename
    shutil.copy(current_path, new_path)
    return new_path



from __future__ import annotations

# The representation, in C++ code, of all the data being passed around the system by the C++
# code. Mostly, these are used as "representations" of a particular AST node. Here is an outline
# of how this works:
#
# Things are broken down, generally, into values and sequences.
#
# A value is something like j->pt(), or perhaps even just j. It has a specific value, and it
# also has a particular place where it is valid - the so-called scope. The scope tells you exactly
# where in the code you are allowed to refer to that value. The value is valid at that scope or
# anything deeper.
#
# A collection is a special type of value. The thing about collection is that you can turn it into
# a sequence. This happens implicitly with code that looks like "e.Jets.Select(xxx)" - between the Jets
# and the Select, the collection that is e.Jets gets turned into a sequence.
#
# A tuple is just a special case of a value, and it contains a series of either values or sequences or collections, etc.
#
# Finally, a sequence is something we are iterating over. Select, SelectMany, and Where all return sequences.
# A sequence has the loop iterator - the thing that we are looping over. This iterator is just a value, and has
# all the properties of a value. A sequence also has a "value" associated with it. This value is at first just
# the iterator. But you can have something like a select statement that will do Select(j->pT()). In that case
# the sequence value transforms from j to j->pT(). THe iterator value remains the same at that point.
#
# It is instructive to talk about how various LINQ predicates transform the above things.
#
# seq.Select(f(x)):
#  - if seq is a collection -> sequence with a new iterator value and the sequence value being f(iterator value)
#  - seq can't be a plain value
#  - if seq is a sequence -> new sequence with same iterator value and sequence value being f(old sequence value).
#    Note it is possible for the sequence value to be another sequence!
#
# seq.SelectMany(f(x)):
#  - if seq is a collection -> sequence with a new iterator value and sequence value being f(iterator value).
#    Then go onto next step that applies.
#  - seq can't be a plane value
#
#  next look a the sequence value:
#  - collection -> start a new sequence with the iterator and sequence value being the iterator
#  - value -> not allowedNone,
#  - sequence -> keep the sequence
#
# seq.First()
#  - seq is a collection -> start a new sequence with the iterator and sequence being the iterator
#  - seq is a value -> not allowed
# then
#  - seq is a sequence -> return the sequence value
#
# Others follow a similar line of reasoning.
import ast
import copy
from typing import Any, Optional, TypeVar, Union, cast

import func_adl_xAOD.common.cpp_types as ctyp
from func_adl_xAOD.common.util_scope import gc_scope, gc_scope_top_level


class dummy_ast(ast.AST):
    "A dummy ast"
    _fields = tuple()

    def __init__(self, node: cpp_rep_base):
        set_rep(self, node)


class cpp_rep_base:
    r"""
    Represents a term or collection in C++ code. Base class for all values, collections, sequences.

    This is an abstract class for the most part. Do not override things that aren't needed - that way the system will
    know when the user tries to do something that they shouldn't have.

    We know only about our type
    """

    def __init__(self):
        pass

    def as_ast(self):
        "Create an AST rep of this guy"
        return dummy_ast(self)


class cpp_namespace(cpp_rep_base):
    def __init__(self, ns: ctyp.NameSpaceInfo):
        """Create a representation of a namespace in C++ code"""
        self._ns = ns

    def __str__(self) -> str:
        return f"cpp_namespace({self._ns.full_name})"

    def scope(self) -> gc_scope_top_level:
        return gc_scope_top_level()

    @property
    def ns(self) -> ctyp.NameSpaceInfo:
        return self._ns


class cpp_enum(cpp_rep_base):
    def __init__(self, enum: ctyp.ENumInfo):
        """Create a representation of a enum in C++ code"""
        self._enum = enum

    def __str__(self) -> str:
        return f"cpp_enum({self._enum.full_name})"

    def scope(self) -> gc_scope_top_level:
        return gc_scope_top_level()

    @property
    def enum(self) -> ctyp.ENumInfo:
        return self._enum


class cpp_value(cpp_rep_base):
    r"""
    Represents a value. This has a particular value in C++ code that is valid at some C++ scope, or deeper.
    """

    def __init__(
        self,
        cpp_expression: str,
        scope: Optional[Union[gc_scope, gc_scope_top_level]],
        cpp_type: Optional[ctyp.terminal],
    ):
        r"""
        Initialize a C++ value

        cpp_expression:         The C++ name that can be emitted to code
        scope:                  The scope at which this value is valid
        cpp_type:               Type that we can deal with
        """
        cpp_rep_base.__init__(self)
        self._scope = scope
        self._expression = cpp_expression
        assert not isinstance(cpp_type, str)
        self._cpp_type = cpp_type

    def __str__(self) -> str:
        return f"{str(self._cpp_type)} value (expression {self._expression})"

    @property
    def p_depth(self) -> int:
        "Return the depth of the pointer"
        return self.cpp_type().p_depth

    def as_cpp(self) -> str:
        return self._expression

    def reset_scope(self, scope: gc_scope):
        "If scope has not been set, then we can set it"
        # TODO: #116 This feels like a kludge - should we pass around something else and get rid of this?
        if self._scope is None:
            self._scope = scope
        else:
            raise RuntimeError(
                "Internal Error: Scope is already set - it can't be set to a new value"
            )

    def scope(self) -> Union[gc_scope, gc_scope_top_level]:
        "Return the scope at which this variable becomes valid."
        if self._scope is not None:
            return self._scope
        else:
            raise RuntimeError(
                "Internal Error: Asking for the undefined scope of a value."
            )

    def cpp_type(self) -> ctyp.terminal:
        if self._cpp_type is None:
            raise RuntimeError(
                f"Internal Error: Variable {self._expression} does not have an assigned type, but needs one."
            )
        return self._cpp_type

    def has_cpp_type(self) -> bool:
        return self._cpp_type is not None

    def copy_with_new_scope(self, scope) -> cpp_value:
        "Make a new version, with just the scope changed"
        new_v = copy.copy(self)
        new_v._scope = scope
        return new_v

    def update_type(self, new_type: ctyp.terminal):
        """
        Update to a new type
        """
        self._cpp_type = new_type


class cpp_variable(cpp_value):
    r"""
    A value that can be declared. Refers to a variable.
    """

    def __init__(
        self,
        cpp_expression: str,
        scope: Union[gc_scope, gc_scope_top_level],
        cpp_type,
        initial_value: Optional[cpp_value] = None,
    ):
        cpp_value.__init__(self, cpp_expression, scope, cpp_type)
        self._initial_value = initial_value

    def initial_value(self):
        return self._initial_value

    def update_type(self, new_type: ctyp.terminal):
        if self._initial_value is not None:
            self._initial_value._cpp_type = new_type
        cpp_value.update_type(self, new_type)


class cpp_collection(cpp_value):
    r"""
    Represents a special kind of value - a collection (vector<float>).
    """

    def __init__(
        self,
        cpp_expression: str,
        scope: Union[gc_scope, gc_scope_top_level],
        collection_type: ctyp.collection,
    ):
        r"""
        Initialize a C++ value that can be turned into a sequence if requested.

        cpp_expression:         The expression in C++ to refer to this collection.
        scope:                  The scope at which this collection is valid.
        collection_type:        The type of the collection (see cpp_types). It should be a collection
        """
        cpp_value.__init__(self, cpp_expression, scope, collection_type)

    def get_element_type(self):
        "Return the type of the element of the sequence"
        return cast(ctyp.collection, self.cpp_type()).element_type

    def token_type(self):
        "Return the type of the token required, such as edm::EDGetTokenT<pat::MuonCollection> for miniAOD file"
        return None


class cpp_tuple(cpp_rep_base):
    r"""
    Represents a special kind of value - a tuple, which is just a container of other values. This
    isn't a regular cpp_value in the sense it can't be directly put into C++ code. It must be
    specially handled by the interpreter.
    """

    def __init__(
        self, list_of_values: tuple, scope: Union[gc_scope, gc_scope_top_level]
    ):
        """
        A Tuple holds a list of different types

        list_of_values:         Iterator that will give us all the cpp_xxx.
        scope:                  Scope where this is valid
        cpp_type:               An instance of cpp_types.tuple to specify the type.
        """
        cpp_rep_base.__init__(self)
        self._values = list_of_values
        self._scope = scope

    def values(self):
        return self._values

    def scope(self):
        return self._scope


class cpp_dict(cpp_rep_base):
    """Represents a special kind of value = a dict, which is just a keyed container of other values.
    This isn't a regular cpp_value in the sense it can't directly be put into C++ code. It must
    be specially handled by the interpreter.
    """

    def __init__(self, value: dict, scope: Union[gc_scope, gc_scope_top_level]):
        super().__init__()
        self._values = value
        self._scope = scope

    @property
    def value_dict(self) -> dict:
        return self._values

    def scope(self) -> Union[gc_scope, gc_scope_top_level]:
        return self._scope


class cpp_sequence(cpp_rep_base):
    """
    Represents an iterator over a sequence of data. The `sequence_value` points to value of the item in
    the stream, and `iterator_value` points to the iterator itself we are looping over (appears in the
    for statement).

    A sequence is a stream of values of a particular type. You can think of it like a generator expression,
    or like a iterator into a C++ vector of some type.
    """

    def __init__(
        self,
        sequence_value: Union[cpp_value, cpp_sequence],
        iterator_value: cpp_value,
        scope: Union[gc_scope_top_level, gc_scope],
    ):
        """
        Create a sequence

        sequence_value:         The value of the sequence - of the data items that are in sequence.
        iterator_value:         The iterator that is incremented to get the next value in the sequence. The
                                iterators scope is defined where it was created.
        scope:                  The scope at which this sequence is declared. Note that this may be different
                                from the of the interator if we are, for example, inside an if statement caused
                                by a Where (or similar). If the `sequence_value` is an actual value, it will have
                                the same scope.
        """
        cpp_rep_base.__init__(self)
        self._sequence = sequence_value
        self._iterator = iterator_value
        self._type: Optional[ctyp.collection] = None
        self._scope = scope

    def sequence_value(self):
        return self._sequence

    def iterator_value(self):
        return self._iterator

    def cpp_type(self) -> ctyp.terminal:
        if self._type is None:
            self._type = ctyp.collection(self._sequence.cpp_type())
        return self._type

    def as_cpp(self):
        raise RuntimeError("Do not know how to get the cpp rep of a sequence!")

    def scope(self) -> Union[gc_scope, gc_scope_top_level]:
        "Return scope where this sequence was created/valid"
        return self._scope


def set_rep(node: ast.AST, value: cpp_rep_base, scope: Optional[Any] = None):
    """
    Set the representation of a node to a value.
    """
    node.rep = value  # type: ignore
    if scope is not None:
        node.scope = scope  # type: ignore


def get_rep(node: ast.AST) -> cpp_rep_base:
    """
    Get the representation of a node.
    """
    return node.rep  # type: ignore


# Allow dereference type for anything that is a value of some sort
DT = TypeVar("DT", bound=cpp_value)


def dereference_var(v: DT) -> DT:
    """
    If this is a pointer, return the object with the proper type (and a * to dereference it). Otherwise
    just return the object itself.

    NOTE: It might just return the object itself, not dereferencing it!
    """
    if not v.cpp_type().is_a_pointer:
        return v

    # We will go under the covers and "fix" this.
    new_v = copy.copy(v)
    new_v._expression = "*" + new_v._expression

    # There is only one type we current support here.
    # Eventually this is going to get us into trouble.
    new_v._cpp_type = new_v.cpp_type().get_dereferenced_type()
    return new_v


def base_type_member_access(v: cpp_value, extra_deref: int = 0) -> str:
    """
    Turn a C++ object into a base reference suitable for
    a member access.

    obj f => f.
    obj *f => f->
    obj **f => (*f)->

    v:            The object to access.
    extra_deref:  The number of extra dereferences to apply.
    """
    result = v.as_cpp()
    depth = extra_deref + v.cpp_type().p_depth
    for _ in range(1, depth):
        result = f"(*{result})"
    if depth > 0:
        return f"{result}->"
    else:
        return f"{result}."



# setuptools loads some plugins necessary for use here.
import os
from distutils.core import setup

from version_info import version_func_adl

# Use the readme as the long description.
with open("README.md", "r") as fh:
    long_description = fh.read()

# Get the version number right
version = os.getenv("func_adl_xaod_version")
if version is None:
    version = "0.0.0-alpha.10"
else:
    version = version.split("/")[-1]

# Finally, the setup.
setup(
    name="func_adl_xAOD",
    version=version,
    packages=["func_adl_xAOD"]
    + [f"func_adl_xAOD.{f}" for f in ["atlas.xaod", "cms.aod", "common"]],
    scripts=[],
    description="Functional Analysis Description Language backend for accessing ATLAS xAOD files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="G. Watts (IRIS-HEP/UW Seattle)",
    author_email="gwatts@uw.edu",
    maintainer="Gordon Watts (IRIS-HEP/UW Seattle)",
    maintainer_email="gwatts@uw.edu",
    url="https://github.com/iris-hep/func_adl_xAOD",
    license="TBD",
    test_suite="tests",
    install_requires=[
        "requests",
        "retry",
        "jinja2",
        "qastle",
        f"func_adl{version_func_adl}",
    ],
    extras_require={
        "test": [
            "pytest>=3.9",
            "pytest-asyncio",
            "pytest-mock",
            "pytest-cov",
            "coverage",
            "flake8",
            "autopep8",
            "twine",
            "testfixtures",
            "wheel",
            "asyncmock",
            "pandas",
            "uproot",
            "awkward",
            "black",
        ],
        "local": ["python-on-whales"],
    },
    classifiers=[
        # "Development Status :: 3 - Alpha",
        # "Development Status :: 4 - Beta",
        "Development Status :: 5 - Production/Stable",
        # "Development Status :: 6 - Mature",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ],
    package_data={
        "func_adl_xAOD": [
            "template/atlas/r21/*",
            "template/cms/r5/*",
        ],
    },
    platforms="Any",
)



# Infrastructure to replace simple functions (like "abs") in python with their
# equivalent in C++.
import ast
from collections import namedtuple
from func_adl_xAOD.common.cpp_types import terminal


class FunctionAST(ast.AST):
    """
    An AST node that represents a function that can be a drop-in replacement for
    a python function
    """

    def __init__(self, cpp_name, include_files, cpp_return_type):
        """Initialize an AST node that represents a C++ drop-in function call

        cpp_name: The C++ name that we will use to do the call
        include_files: List of include files to be put at the start of the emitted source.
        """
        self.cpp_name = cpp_name
        self.include_files = include_files
        self.cpp_return_type = cpp_return_type
        self.fields = []


class find_known_functions(ast.NodeTransformer):
    def visit_Call(self, node):
        "Look for a call to a Name function that is in our list"
        # First go one level down.
        self.generic_visit(node)

        # See if we are candidate for replacement
        # Get a fully qualified name as possible.
        if type(node.func) is not ast.Name:
            return node

        try:
            fnc = eval(node.func.id)
            fnc_name = f"{fnc.__module__}.{node.func.id}"
        except NameError:
            fnc_name = node.func.id

        if fnc_name not in functions_to_replace:
            return node

        # Build the replacement.
        info = functions_to_replace[fnc_name]
        node.func = FunctionAST(info.cpp_name, info.include_files, info.cpp_return_type)  # type: ignore

        return node


def add_function_mapping(python_name, cpp_name, include_files, return_type):
    """Add a re-mapping from a python function to an actual function.

    python_name: fully qualified name of the python function
    cpp_name: fully qualified name of the C++ function
    include_files: any include files that should be included in the C++ source. Can also be a single string.
    return_type: C++ return type
    """
    global functions_to_replace
    functions_to_replace[python_name] = cpp_function(
        cpp_name,
        include_files
        if type(include_files) is list
        else [
            include_files,
        ],
        terminal(return_type),
    )


# The list of functions to do the replacement
functions_to_replace = {}


# The named tuple that stores the replacement info.
# TODO: change to a typed data class
cpp_function = namedtuple(
    "cpp_function", ["cpp_name", "include_files", "cpp_return_type"]
)

# CMATH functions - the catalog was pulled from
# http://www.cplusplus.com/reference/cmath/
# Trig
add_function_mapping("sin", "std::sin", "cmath", "double")
add_function_mapping("cos", "std::cos", "cmath", "double")
add_function_mapping("tan", "std::tan", "cmath", "double")
add_function_mapping("acos", "std::acos", "cmath", "double")
add_function_mapping("asin", "std::asin", "cmath", "double")
add_function_mapping("atan", "std::atan", "cmath", "double")
add_function_mapping("atan2", "std::atan2", "cmath", "double")

# Hyperbolic
add_function_mapping("sinh", "std::sinh", "cmath", "double")
add_function_mapping("cosh", "std::cosh", "cmath", "double")
add_function_mapping("tanh", "std::tanh", "cmath", "double")
add_function_mapping("asinh", "std::asinh", "cmath", "double")
add_function_mapping("acosh", "std::acosh", "cmath", "double")
add_function_mapping("atanh", "std::atanh", "cmath", "double")

# Exponential and Log functions
add_function_mapping("exp", "std::exp", "cmath", "double")
# add_function_mapping('frexp', 'std::frexp', 'cmath', 'double')  # By value parameter
add_function_mapping("ldexp", "std::ldexp", "cmath", "double")
add_function_mapping("log", "std::log", "cmath", "double")
add_function_mapping("ln", "std::log", "cmath", "double")
add_function_mapping("log10", "std::log10", "cmath", "double")
# add_function_mapping('modf', 'std::modf', 'cmath', 'double')  # By value parameter
add_function_mapping("exp2", "std::exp2", "cmath", "double")
add_function_mapping("expm1", "std::expm1", "cmath", "double")
add_function_mapping("ilogb", "std::ilogb", "cmath", "double")
add_function_mapping("log1p", "std::log1p", "cmath", "double")
add_function_mapping("log2", "std::log2", "cmath", "double")
add_function_mapping("scalbn", "std::scalbn", "cmath", "double")
add_function_mapping("scalbln", "std::scalbln", "cmath", "double")

# Power Functions
add_function_mapping("pow", "std::pow", "cmath", "double")
add_function_mapping("sqrt", "std::sqrt", "cmath", "double")
add_function_mapping("cbrt", "std::cbrt", "cmath", "double")
add_function_mapping("hypot", "std::hypot", "cmath", "double")

# Error and Gamma Functions
add_function_mapping("erf", "std::erf", "cmath", "double")
add_function_mapping("erfc", "std::erfc", "cmath", "double")
add_function_mapping("tgamma", "std::tgamma", "cmath", "double")
add_function_mapping("lgamma", "std::lgamma", "cmath", "double")

# Rounding and remainder functions
add_function_mapping("ceil", "std::ceil", "cmath", "double")
add_function_mapping("floor", "std::ceil", "cmath", "double")
add_function_mapping("fmod", "std::ceil", "cmath", "double")
add_function_mapping("trunc", "std::ceil", "cmath", "double")
add_function_mapping("round", "std::ceil", "cmath", "double")
# add_function_mapping('lround', 'std::ceil', 'cmath', 'double')  # Long int
# add_function_mapping('llround', 'std::ceil', 'cmath', 'double')  # Long int
add_function_mapping("rint", "std::ceil", "cmath", "double")
# add_function_mapping('lrint', 'std::ceil', 'cmath', 'double')  # Long int
# add_function_mapping('llrint', 'std::ceil', 'cmath', 'double')  # Long int
add_function_mapping("nearbyint", "std::ceil", "cmath", "double")
add_function_mapping("remainder", "std::ceil", "cmath", "double")
add_function_mapping("remquo", "std::ceil", "cmath", "double")

# Floating-point manipulation functions
add_function_mapping("copysign", "std::copysign", "cmath", "double")
add_function_mapping("nan", "std::nan", "cmath", "double")
add_function_mapping("nextafter", "std::nextafter", "cmath", "double")
add_function_mapping("nexttoward", "std::nexttoward", "cmath", "double")

# Minimum, maximum, difference functions
add_function_mapping("fdim", "std::fdim", "cmath", "double")
add_function_mapping("fmax", "std::fmax", "cmath", "double")
add_function_mapping("fmin", "std::fmin", "cmath", "double")

# Other functions
add_function_mapping("fabs", "std::fabs", "cmath", "double")
add_function_mapping("abs", "std::fabs", "cmath", "double")
add_function_mapping("fma", "std::fma", "cmath", "double")

# Python built-in functions. These are weird as they get a different name, and
# don't really come in as a function.
add_function_mapping("builtins.abs", "std::abs", "cmath", "double")
add_function_mapping(
    "builtins.pow", "std::pow", "cmath", "double"
)  # Not the 3 arg version!

# add_function_mapping('builtins.max', 'std::abs', 'cmath', 'double')  # Do not use to do a sequence
# add_function_mapping('builtins.min', 'std::abs', 'cmath', 'double')  # Do not use to do a sequence
